#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>
#include <gdk/gdkrgb.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "sfm.h"
#include "blitting.h"
#include "guiutils.h"
#include "csd.h"
#include "toolbar.h"
#include "statusbar.h"
#include "editclist.h"

#include "editor.h"
#include "editorcb.h"
#include "editorobj.h"
#include "editorfio.h"
#include "editorscnopen.h"
#include "editorview.h"
#include "scnedit.h"
#include "config.h"

#include "images/scnedit_48x48.xpm"

#include "images/icon_new_20x20.xpm"
#include "images/icon_open_20x20.xpm"
#include "images/icon_save_20x20.xpm"
#include "images/icon_save_as_20x20.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/icon_exit_20x20.xpm"
#include "images/icon_cut_20x20.xpm"
#include "images/icon_copy_20x20.xpm"
#include "images/icon_paste_20x20.xpm"
#include "images/icon_goto_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"
#include "images/icon_reload_20x20.xpm"
#include "images/icon_stop_20x20.xpm"
#include "images/icon_select_20x20.xpm"
#include "images/icon_run_20x20.xpm"

#include "images/icon_object_32x32.xpm"

#include "images/icon_object_20x20.xpm"
#include "images/icon_ground_20x20.xpm"
#include "images/icon_helipad_20x20.xpm"
#include "images/icon_runway_20x20.xpm"
#include "images/icon_human_20x20.xpm"
/*
#include "images/icon_fire_20x20.xpm"
#include "images/icon_smoke_20x20.xpm"
 */

#include "images/icon_object_15x15.xpm"
#include "images/icon_ground_15x15.xpm"
#include "images/icon_helipad_15x15.xpm"
#include "images/icon_human_15x15.xpm"
#include "images/icon_fire_15x15.xpm"
#include "images/icon_smoke_15x15.xpm"


gint EditorGetObjType(GtkWidget *w);
void EditorSetObjType(GtkWidget *w, gint type);
gint EditorGetSizeShape(GtkWidget *w);
void EditorSetSizeShape(GtkWidget *w, gint shape);
gint EditorGetRunwaySurface(GtkWidget *w);
void EditorSetRunwaySurface(GtkWidget *w, gint surface);
gboolean EditorIsObjectNameSpecial(const gchar *name);
gboolean EditorGetRegisteredLocation(
        editor_struct *e, gint i,
        gfloat *xm, gfloat *ym, gfloat *zm,
        gfloat *heading, gfloat *pitch, gfloat *bank,
        gchar **name
);

void EditorHdrMapDraw(editor_struct *e);

static void EditorSetDrawingAreaEvents(
	GtkWidget *w, gpointer func, gpointer data
);
static void EditorBuildMenuBar(
	editor_struct *e, GtkWidget *parent, GdkWindow *window,
	const options_struct *opt
);
static void EditorBuildToolBar(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
);
static void EditorBuildHeaderPage(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
);
static void EditorBuildObjectValues(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
);
static void EditorBuildSceneryPage(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
);
editor_struct *EditorNew(gpointer core_ptr);
void EditorUpdateTitle(editor_struct *e);
void EditorUpdateMenus(editor_struct *e);
void EditorSavePosition(editor_struct *e);
void EditorSetBusy(editor_struct *e, gboolean is_busy);
void EditorStopReset(editor_struct *e, gboolean sensitive);
void EditorMap(editor_struct *e);
void EditorUnmap(editor_struct *e);
void EditorDelete(editor_struct *e);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? (gfloat)atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)

#define REMOVE_TOID(i)  { if(*(i) != (guint)-1) { \
 gtk_timeout_remove(*(i)); *(i) = (guint)-1; } }


/*
 *	Returns one of SAR_OBJ_TYPE_* based on the current value of the
 *	given GtkCombo w.
 */
gint EditorGetObjType(GtkWidget *w)
{
	GtkCombo *combo = (GtkCombo *)w;
	GtkEntry *entry = (combo != NULL) ? GTK_ENTRY(combo->entry) : NULL;
	const gchar *s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;

	if(s == NULL)
	    return(SAR_OBJ_TYPE_STATIC);

	if(!g_strcasecmp(s, "static"))
	    return(SAR_OBJ_TYPE_STATIC);
        else if(!g_strcasecmp(s, "ground"))
            return(SAR_OBJ_TYPE_GROUND);
        else if(!g_strcasecmp(s, "runway"))
            return(SAR_OBJ_TYPE_RUNWAY);
        else if(!g_strcasecmp(s, "helipad"))
            return(SAR_OBJ_TYPE_HELIPAD);
        else if(!g_strcasecmp(s, "human"))
            return(SAR_OBJ_TYPE_HUMAN);
        else if(!g_strcasecmp(s, "smoke"))
            return(SAR_OBJ_TYPE_SMOKE);
        else if(!g_strcasecmp(s, "fire"))
            return(SAR_OBJ_TYPE_FIRE);
        else if(!g_strcasecmp(s, "premodeled"))
            return(SAR_OBJ_TYPE_PREMODELED);
	else
	    return(SAR_OBJ_TYPE_STATIC);
}

/*
 *	Sets the object type for the given GtkCombo w, the type must be
 *	one of SAR_OBJ_TYPE_*.
 */
void EditorSetObjType(GtkWidget *w, gint type)
{
	const gchar *s;
        GtkCombo *combo = (GtkCombo *)w;
        GtkEntry *entry = (combo != NULL) ? GTK_ENTRY(combo->entry) : NULL;

	if(entry == NULL)
	    return;

	switch(type)
	{
	  case SAR_OBJ_TYPE_STATIC: s = "Static"; break;
          case SAR_OBJ_TYPE_GROUND: s = "Ground"; break;
          case SAR_OBJ_TYPE_RUNWAY: s = "Runway"; break;
          case SAR_OBJ_TYPE_HELIPAD: s = "Helipad"; break;
          case SAR_OBJ_TYPE_HUMAN: s = "Human"; break;
          case SAR_OBJ_TYPE_SMOKE: s = "Smoke"; break;
          case SAR_OBJ_TYPE_FIRE: s = "Fire"; break;
          case SAR_OBJ_TYPE_PREMODELED: s = "Premodeled"; break;
	  default: s = "Static"; break;
	}
	gtk_entry_set_text(entry, s);
}

/*
 *	Returns one of SAR_CONTACT_SHAPE_* based on the current value of
 *	the given GtkCombo w.
 */
gint EditorGetSizeShape(GtkWidget *w)
{
        GtkCombo *combo = (GtkCombo *)w;
        GtkEntry *entry = (combo != NULL) ? GTK_ENTRY(combo->entry) : NULL;
        const gchar *s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;

        if(s == NULL)
            return(SAR_CONTACT_SHAPE_SPHERICAL);

        if(!g_strcasecmp(s, "Spherical"))
            return(SAR_CONTACT_SHAPE_SPHERICAL);
        else if(!g_strcasecmp(s, "Cylendrical"))
            return(SAR_CONTACT_SHAPE_CYLENDRICAL);
        else if(!g_strcasecmp(s, "Rectangular"))
            return(SAR_CONTACT_SHAPE_RECTANGULAR);
	else
	    return(SAR_CONTACT_SHAPE_SPHERICAL);
}

/*
 *      Sets the contact shape for the given GtkCombo w, the type must
 *	be one of SAR_CONTACT_SHAPE_*.
 */
void EditorSetSizeShape(GtkWidget *w, gint shape)
{
        const gchar *s;
        GtkCombo *combo = (GtkCombo *)w;
        GtkEntry *entry = (combo != NULL) ? GTK_ENTRY(combo->entry) : NULL;

        if(entry == NULL)
            return;

        switch(shape)
        {
          case SAR_CONTACT_SHAPE_SPHERICAL: s = "Spherical"; break;
          case SAR_CONTACT_SHAPE_CYLENDRICAL: s = "Cylendrical"; break;
          case SAR_CONTACT_SHAPE_RECTANGULAR: s = "Rectangular"; break;
          default: s = "Spherical"; break;
        }
        gtk_entry_set_text(entry, s);
}

/*
 *	Returns one of SAR_RUNWAY_SURFACE_* based on the current value
 *	of the given GtkCombo w.
 */
gint EditorGetRunwaySurface(GtkWidget *w)
{
        GtkCombo *combo = (GtkCombo *)w;
        GtkEntry *entry = (combo != NULL) ? GTK_ENTRY(combo->entry) : NULL;
        const gchar *s = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;

        if(s == NULL)
            return(SAR_RUNWAY_SURFACE_PAVED);

        if(!g_strcasecmp(s, "Paved"))
            return(SAR_RUNWAY_SURFACE_PAVED);
        else if(!g_strcasecmp(s, "Gravel"))
            return(SAR_RUNWAY_SURFACE_GRAVEL);
        else if(!g_strcasecmp(s, "Concrete"))
            return(SAR_RUNWAY_SURFACE_CONCRETE);
        else if(!g_strcasecmp(s, "Groved"))
            return(SAR_RUNWAY_SURFACE_GROVED);
        else
            return(SAR_RUNWAY_SURFACE_PAVED);
}

/*
 *	Sets the runway surface for the given GtkCombo w, the surface
 *	must be one of SAR_RUNWAY_SURFACE_*.
 */
void EditorSetRunwaySurface(GtkWidget *w, gint surface)
{
        const gchar *s;
        GtkCombo *combo = (GtkCombo *)w;
        GtkEntry *entry = (combo != NULL) ? GTK_ENTRY(combo->entry) : NULL;

        if(entry == NULL)
            return;

        switch(surface)
        {
          case SAR_RUNWAY_SURFACE_PAVED: s = "Paved"; break;
          case SAR_RUNWAY_SURFACE_GRAVEL: s = "Gravel"; break;
          case SAR_RUNWAY_SURFACE_CONCRETE: s = "Concrete"; break;
          case SAR_RUNWAY_SURFACE_GROVED: s = "Groved"; break;
	  default: s = "Paved"; break;
        }
        gtk_entry_set_text(entry, s);
}

/*
 *	Returns TRUE if the given name matches one of the special
 *	object names.
 */
gboolean EditorIsObjectNameSpecial(const gchar *name)
{
	gint i = 0;
	gchar *names[] = SE_SAR_SPECIAL_OBJECT_NAMES;

	if(name == NULL)
	    return(FALSE);

	for(i = 0; names[i] != NULL; i++)
	{
	    if(!g_strcasecmp(names[i], name))
		return(TRUE);
	}

	return(FALSE);
}

/*
 *	Returns the position, direction, and name of the registered 
 *	location specified by i.
 *
 *	Returns TRUE if a registered location specified by i exists or
 *	FALSE on error.
 */
gboolean EditorGetRegisteredLocation(
        editor_struct *e, gint i,
        gfloat *xm, gfloat *ym, gfloat *zm,
        gfloat *heading, gfloat *pitch, gfloat *bank,
	gchar **name
)
{
	gint row, column;
	gchar *s = NULL;
	editclist_struct *eclist = (e != NULL) ?
	    e->hdr_registered_locations_eclist : NULL;
	GtkCList *clist = (GtkCList *)EDITCLIST_CLIST(eclist);

	if(xm != NULL)
	    *xm = 0.0f;
        if(ym != NULL)
            *ym = 0.0f;
        if(zm != NULL)
            *zm = 0.0f;
        if(heading != NULL)
            *heading = 0.0f;
        if(pitch != NULL)
            *pitch = 0.0f;
        if(bank != NULL)
            *bank = 0.0f;
	if(name != NULL)
	    *name = NULL;

	if(clist == NULL)
	    return(FALSE);

	if((i < 0) || (i >= clist->rows))
	    return(FALSE);

	row = i;

	/* Get name. */
	column = 0;
	if(name != NULL)
	    gtk_clist_get_text(clist, row, column, name);

	/* Get position. */
        column = 1;
        if(xm != NULL)
	{
            gtk_clist_get_text(clist, row, column, &s);
	    *xm = ATOF(s);
	}
        column = 2;
        if(ym != NULL)
        {
            gtk_clist_get_text(clist, row, column, &s);
            *ym = ATOF(s);
        }
        column = 3;
        if(zm != NULL)
        {
            gtk_clist_get_text(clist, row, column, &s);
            *zm = (gfloat)SFMFeetToMeters(ATOF(s));
        }

	/* Get direction. */
        column = 4;
        if(heading != NULL)
        {
            gtk_clist_get_text(clist, row, column, &s);
            *heading = (gfloat)DEGTORAD(ATOF(s));
        }
        column = 5;
        if(pitch != NULL)
        {
            gtk_clist_get_text(clist, row, column, &s);
            *pitch = (gfloat)DEGTORAD(ATOF(s));
        }
        column = 6;
        if(bank != NULL)
        {
            gtk_clist_get_text(clist, row, column, &s);
            *bank = (gfloat)DEGTORAD(ATOF(s));
        }

	return(TRUE);
}


/*
 *	Redraws the map.
 */
void EditorHdrMapDraw(editor_struct *e)
{
	gint state, width, height;
	gfloat zoom;
	const editor_tex_struct *t;
	GdkWindow *window;
	GdkGC *gc;
	GtkStyle *style;
	GtkWidget *w;


	if(e == NULL)
	    return;

	t = e->hdr_map_tex;
	w = e->hdr_map_da;
	if((t == NULL) || (w == NULL))
	    return;

	if(!GTK_WIDGET_REALIZED(w))
	    gtk_widget_realize(w);

	state = GTK_WIDGET_STATE(w);
	window = w->window;
	if((window == NULL) || (e->hdr_map_buf == NULL))
	    return;

	style = gtk_widget_get_style(w);
	if(style == NULL)
	    return;

	gc = style->fg_gc[state];
	gdk_window_get_size(window, &width, &height);
	if((width <= 0) || (height <= 0) ||
           (t->width <= 0) || (t->height <= 0)
	)
	    return;

	zoom = (gfloat)width / (gfloat)t->width;
	BlitImageScaleRGBA(
	    BLIT_PIXEL_OP_SET,
	    (guint32 *)e->hdr_map_buf,
	    0, 0, width, height, -1,
	    (const guint32 *)t->data,
	    0, 0, t->width, t->height, t->bpl,
	    zoom, zoom
	);
	gdk_draw_rgb_32_image(
	    window, gc,
	    0, 0, width, height,
	    GDK_RGB_DITHER_NORMAL,
	    (guchar *)e->hdr_map_buf,
	    width * 4
	);

        /* Draw focus? */
        if(GTK_WIDGET_HAS_FOCUS(w) && GTK_WIDGET_SENSITIVE(w))
            gdk_draw_rectangle(
                window, gc, FALSE,
                0, 0, width - 1, height - 1
            );
}


/*
 *	Sets up standard events for the GtkDrawingArea widget.
 */
static void EditorSetDrawingAreaEvents(
        GtkWidget *w, gpointer func, gpointer data
)
{
	if(w == NULL)
	    return;

	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
        gtk_widget_add_events(
            w,
            GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
            GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
            GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
            GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK |
            GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
            GDK_FOCUS_CHANGE_MASK
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "configure_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "expose_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_press_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "motion_notify_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "enter_notify_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "leave_notify_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "focus_in_event",
            GTK_SIGNAL_FUNC(func), data
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "focus_out_event",
            GTK_SIGNAL_FUNC(func), data
        );
}

/*
 *      Builds the menu bar.
 */
static void EditorBuildMenuBar(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
)
{
	guint8 **icon;
	const gchar *label;
	gint accel_key;
	guint accel_mods;
	gpointer mclient_data = e;
	void (*func_cb)(GtkWidget *w, gpointer);
	gpointer fw;
	GtkAccelGroup *accelgrp = e->accelgrp;
        GtkWidget *menu_bar, *menu, *w;

        /* Create menu bar. */
	e->menu_bar = menu_bar = (GtkWidget *)GUIMenuBarCreate(NULL);
        if(GTK_IS_BOX(parent))
            gtk_box_pack_start(GTK_BOX(parent), menu_bar, FALSE, FALSE, 0);
        else
            gtk_container_add(GTK_CONTAINER(parent), menu_bar);
        gtk_widget_show(menu_bar);

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (gpointer *)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_ITEM_CHECK  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp, \
  icon, label, accel_key, accel_mods, (gpointer *)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }

        /* Create file menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            icon = (guint8 **)icon_new_20x20_xpm;
            label = "New";
            accel_key = 'n';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorNewCB;
            DO_ADD_MENU_ITEM_LABEL
	    e->new_mi = w;

            icon = (guint8 **)icon_open_20x20_xpm;
            label = "Open...";
            accel_key = 'o';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorOpenCB;
            DO_ADD_MENU_ITEM_LABEL
	    e->open_mi = w;

            icon = (guint8 **)icon_save_20x20_xpm;
            label = "Save";
            accel_key = 's';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorSaveCB;
            DO_ADD_MENU_ITEM_LABEL
	    e->save_mi = w;

            icon = (guint8 **)icon_save_as_20x20_xpm;
            label = "Save As...";
            accel_key = 'a';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorSaveAsCB;
            DO_ADD_MENU_ITEM_LABEL
            e->save_as_mi = w;

	    DO_ADD_MENU_SEP

            icon = (guint8 **)icon_exit_20x20_xpm;
            label = "Exit";
            accel_key = 'w';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorCloseCB;
            DO_ADD_MENU_ITEM_LABEL
	    e->exit_mi = w;
	}
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "File",
            GUI_MENU_BAR_ALIGN_LEFT
        );

        /* Create edit menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            icon = (guint8 **)icon_cut_20x20_xpm;
            label = "Cut";
            accel_key = 'x';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorCutCB;
            DO_ADD_MENU_ITEM_LABEL
            e->cut_mi = w;

            icon = (guint8 **)icon_copy_20x20_xpm;
            label = "Copy";
            accel_key = 'c';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorCopyCB;
            DO_ADD_MENU_ITEM_LABEL
            e->copy_mi = w;

            icon = (guint8 **)icon_paste_20x20_xpm;
            label = "Paste";
            accel_key = 'v';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorPasteCB;
            DO_ADD_MENU_ITEM_LABEL
            e->paste_mi = w;

            DO_ADD_MENU_SEP

            icon = (guint8 **)icon_object_20x20_xpm;
            label = "New Object";
            accel_key = GDK_Insert;
            accel_mods = 0;
            func_cb = EditorNewObjectCB;
            DO_ADD_MENU_ITEM_LABEL
            e->new_object_mi = w;

            icon = (guint8 **)icon_goto_20x20_xpm;
            label = "GoTo Object";
            accel_key = 'g';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorGoToSelectedCB;
            DO_ADD_MENU_ITEM_LABEL
            e->goto_selected_object_mi = w;

            icon = NULL;
            label = "Move Object To Cursor";
            accel_key = GDK_F7;
            accel_mods = 0;
            func_cb = EditorMoveToCursorCB;
            DO_ADD_MENU_ITEM_LABEL
            e->move_object_to_cursor_mi = w;

            icon = (guint8 **)icon_cancel_20x20_xpm;
            label = "Delete Object";
            accel_key = GDK_Delete;
            accel_mods = 0;
            func_cb = EditorDeleteObjectCB;
            DO_ADD_MENU_ITEM_LABEL
            e->delete_object_mi = w;
	}
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "Edit",
            GUI_MENU_BAR_ALIGN_LEFT
        );

        /* Create view menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            icon = NULL;
            label = "Header";
            accel_key = '1';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorViewHeaderCB;
            DO_ADD_MENU_ITEM_CHECK
            e->view_header_micheck = w;

            icon = NULL;
            label = "Scenery";
            accel_key = '2';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = EditorViewSceneryCB;
            DO_ADD_MENU_ITEM_CHECK
            e->view_scenery_micheck = w;

	    DO_ADD_MENU_SEP

            icon = NULL;
            label = "Show Map";
            accel_key = GDK_F9;
            accel_mods = 0;
            func_cb = EditorViewDisplayMapCB;
            DO_ADD_MENU_ITEM_CHECK
            e->view_map_micheck = w;

            icon = NULL;
            label = "Show Object Labels";
            accel_key = GDK_F10;
            accel_mods = 0;
            func_cb = EditorViewDisplayObjectLabelsCB;
            DO_ADD_MENU_ITEM_CHECK
            e->view_object_labels_micheck = w;

            icon = NULL;
            label = "Show Position Label";
            accel_key = GDK_F11;
            accel_mods = 0;
            func_cb = EditorViewDisplayPositionLabelCB;
            DO_ADD_MENU_ITEM_CHECK
            e->view_position_label_micheck = w;

            icon = NULL;
            label = "Show Registered Locations";
            accel_key = GDK_F12;
            accel_mods = 0;
            func_cb = EditorViewDisplayRegisteredLocationsCB;
            DO_ADD_MENU_ITEM_CHECK
            e->view_registered_locations_micheck = w;

	    DO_ADD_MENU_SEP

            icon = (guint8 **)icon_reload_20x20_xpm;
            label = "Refresh";
            accel_key = GDK_F5;
            accel_mods = 0;
            func_cb = EditorRefreshCB;
            DO_ADD_MENU_ITEM_LABEL
	    e->refresh_mi = w;
	}
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "View",
            GUI_MENU_BAR_ALIGN_LEFT
        );

        /* Create tools menu. */
        menu = GUIMenuCreateTearOff();
        if(menu != NULL)
        {
            icon = (guint8 **)icon_select_20x20_xpm;
            label = "Check Scenery";
            accel_key = GDK_F6;
            accel_mods = 0;
            func_cb = EditorCheckSceneryCB;
            DO_ADD_MENU_ITEM_LABEL
            e->check_scenery_mi = w;

	    DO_ADD_MENU_SEP

            icon = (guint8 **)icon_run_20x20_xpm;
            label = "Run Search and Rescue";
            accel_key = GDK_F8;
            accel_mods = 0;
            func_cb = EditorRunSARCB;
            DO_ADD_MENU_ITEM_LABEL
            e->run_sar_mi = w;
        }
        GUIMenuAddToMenuBar(
            menu_bar, menu,
            "Tools",
            GUI_MENU_BAR_ALIGN_LEFT
        );



#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_ITEM_CHECK
#undef DO_ADD_MENU_SEP
}

/*
 *	Builds the tool bar.
 */
static void EditorBuildToolBar(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
)
{
	editor_opid_struct *opid;
        toolbar_struct *tool_bar;
        toolbar_item_struct **item = NULL, *item_ptr;
        gint i, total_items = 0;

        gint type;
        const gchar *text;
        guint8 **icon_data;
        const gchar *tooltip;
        gint id;
        void (*func_cb)(gpointer) = EditorToolBarCB;
        void (*enter_cb)(gpointer) = EditorToolBarEnterCB;
        void (*leave_cb)(gpointer) = EditorToolBarLeaveCB;
	void (*mfunc_cb)(GtkWidget *, gpointer);
	gboolean append_to_toolbar;

#define TOOLBAR_ITEM_APPEND	\
{ \
 /* Append new opid. */ \
 i = e->total_opids; e->total_opids = i + 1; \
 e->opid = (editor_opid_struct **)g_realloc( \
  e->opid,  e->total_opids * sizeof(editor_opid_struct *) \
 ); \
 e->opid[i] = opid = EDITOR_OPID(g_malloc0( \
  sizeof(editor_opid_struct) \
 )); \
 opid->editor = e; \
 opid->id = id; \
 opid->tooltip = (tooltip != NULL) ? g_strdup(tooltip) : NULL; \
 opid->func_cb = mfunc_cb; \
 \
 /* Append new tool bar item. */ \
 if(append_to_toolbar) { \
  i = total_items; total_items = i + 1; \
  item = (toolbar_item_struct **)g_realloc( \
   item, total_items * sizeof(toolbar_item_struct *) \
  ); \
  item[i] = item_ptr = ToolBarItemNew( \
   type, text, icon_data, tooltip, id, \
   func_cb, opid, \
   enter_cb, opid, \
   leave_cb, opid \
  ); \
 } \
}

        type = TOOLBAR_ITEM_BUTTON;
        text = "New";
        icon_data = (guint8 **)icon_new_20x20_xpm;
        tooltip = "Create a new scenery";
        id = EDITOR_OPID_NEW;
        mfunc_cb = EditorNewCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Open";
        icon_data = (guint8 **)icon_open_20x20_xpm;
        tooltip = "Open scenery file";
        id = EDITOR_OPID_OPEN;
        mfunc_cb = EditorOpenCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Save";
        icon_data = (guint8 **)icon_save_20x20_xpm;
        tooltip = "Save scenery file";
        id = EDITOR_OPID_SAVE;
        mfunc_cb = EditorSaveCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Save As";
        icon_data = (guint8 **)icon_save_as_20x20_xpm;
        tooltip = "Save scenery to a new file name";
        id = EDITOR_OPID_SAVE_AS;
        mfunc_cb = EditorSaveAsCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Close";
        icon_data = (guint8 **)icon_close_20x20_xpm;
        tooltip = "Close this window";
        id = EDITOR_OPID_CLOSE;
        mfunc_cb = EditorCloseCB;
        append_to_toolbar = FALSE;
        TOOLBAR_ITEM_APPEND

	type = TOOLBAR_ITEM_BUTTON;
	text = "Exit";
	icon_data = (guint8 **)icon_exit_20x20_xpm;
	tooltip = "Close all windows of this application and exit";
	id = EDITOR_OPID_EXIT;
	mfunc_cb = EditorCloseCB;
	append_to_toolbar = FALSE;
	TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_SEPARATOR;
        text = NULL;
        icon_data = NULL;
        tooltip = NULL;
        id = 0;
        mfunc_cb = NULL;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Cut";
        icon_data = (guint8 **)icon_cut_20x20_xpm;
        tooltip = "Cut selected object(s)";
        id = EDITOR_OPID_CUT;
        mfunc_cb = EditorCutCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Copy";
        icon_data = (guint8 **)icon_copy_20x20_xpm;
        tooltip = "Copy selected object(s)";
        id = EDITOR_OPID_COPY;
        mfunc_cb = EditorCopyCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Paste";
        icon_data = (guint8 **)icon_paste_20x20_xpm;
        tooltip = "Paste object(s)";
        id = EDITOR_OPID_PASTE;
        mfunc_cb = EditorPasteCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_SEPARATOR;
        text = NULL;
        icon_data = NULL;
        tooltip = NULL;
        id = 0;
        mfunc_cb = NULL;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "New Obj";
        icon_data = (guint8 **)icon_object_20x20_xpm;
        tooltip = "Create a new object";
        id = EDITOR_OPID_NEW_OBJECT;
        mfunc_cb = EditorNewObjectCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Goto Obj";
        icon_data = (guint8 **)icon_goto_20x20_xpm;
        tooltip = "Go to selected object";
        id = EDITOR_OPID_GOTO_OBJECT;
        mfunc_cb = EditorGoToSelectedCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Del Obj";
        icon_data = (guint8 **)icon_cancel_20x20_xpm;
        tooltip = "Delete selected object(s)";
        id = EDITOR_OPID_GOTO_OBJECT;
        mfunc_cb = EditorDeleteObjectCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_SEPARATOR;
        text = NULL;
        icon_data = NULL;
        tooltip = NULL;
        id = 0;
        mfunc_cb = NULL;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Refresh";
        icon_data = (guint8 **)icon_reload_20x20_xpm;
        tooltip = "Refresh the scenery";
        id = EDITOR_OPID_REFRESH;
        mfunc_cb = EditorRefreshCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

        type = TOOLBAR_ITEM_BUTTON;
        text = "Stop";
        icon_data = (guint8 **)icon_stop_20x20_xpm;
        tooltip = "Stop the current operation";
        id = EDITOR_OPID_STOP;
        mfunc_cb = EditorStopCB;
        append_to_toolbar = TRUE;
        TOOLBAR_ITEM_APPEND

#undef TOOLBAR_ITEM_APPEND

	/* Create new tool bar. */
	if(GTK_IS_BOX(parent))
	{
	    GtkWidget *w = gtk_frame_new(FALSE);
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent = w;
	}
	e->tool_bar = tool_bar = ToolBarNew(
	    item, total_items, parent,
	    TOOLBAR_DISPLAY_PICTURES_AND_TEXT,
	    TOOLBAR_RELIEF_NONE,
	    FALSE
	);
	ToolBarMap(tool_bar);

	/* Delete all tool bar items. */
	for(i = 0; i < total_items; i++)
	    ToolBarItemDelete(item[i]);
	g_free(item);
}

/*
 *	Called by EditorNew() to build the header widgets.
 */
static void EditorBuildHeaderPage(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
)
{
        gint border_major = 5, border_minor = 2;
	gchar *heading[7];
        gpointer entry_rtn;
        GtkAdjustment *adj;
        GtkWidget *w, *parent2, *parent3, *parent4, *parent5;
        GtkEntry *entry;
	GtkCList *clist;
	editclist_struct *eclist;

#define SET_EDITABLE_SIGNAL_CHANGED			\
{ if((w != NULL) ? GTK_IS_EDITABLE(w) : FALSE) {	\
 gtk_signal_connect(					\
  GTK_OBJECT(w), "changed",				\
  GTK_SIGNAL_FUNC(EditorHdrEditableChangedCB), e	\
 );							\
} }
#define SET_TOGGLE_BUTTON_SIGNAL_CHANGED		\
{ if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE) {	\
 gtk_signal_connect(					\
  GTK_OBJECT(w), "toggled",				\
  GTK_SIGNAL_FUNC(EditorHdrToggleButtonToggledCB), e	\
 );							\
} }

        /* Scrolled window for header widgets. */
        e->hdr_sw = w = gtk_scrolled_window_new(NULL, NULL);
        gtk_scrolled_window_set_policy(
            GTK_SCROLLED_WINDOW(w),
            GTK_POLICY_AUTOMATIC,
            GTK_POLICY_AUTOMATIC
        );
        gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent2 = w;
        /* Event box. */
        e->hdr_sw_eb = w = gtk_event_box_new();
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
        gtk_widget_add_events(
            w,
            GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
            GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
            GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_press_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "motion_notify_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_scrolled_window_add_with_viewport(
            GTK_SCROLLED_WINDOW(parent2), w
        );
        gtk_widget_show(w);
        parent2 = w;
	/* Vbox. */
        w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent2), w);
        gtk_widget_show(w);
        parent2 = w;


        /* Locale frame. */
        w = gtk_frame_new("Locale");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* Name. */
	w = GUIPromptBar(NULL, "Name:", NULL, &entry_rtn);
	entry = (GtkEntry *)entry_rtn;
	e->hdr_name_entry = GTK_WIDGET(entry);
	gtk_widget_set_usize(GTK_WIDGET(entry), 300, -1);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
	    GTK_WIDGET(entry),
"Enter the name of the scenery"
        );
        gtk_widget_show(w);
	w = GTK_WIDGET(entry);
	SET_EDITABLE_SIGNAL_CHANGED

	/* Longitude. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Longitude:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -140.0f, 140.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_gps_longitude_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 70, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set longitude center of the scenery (in degrees)"
        );
        gtk_widget_show(w);
	SET_EDITABLE_SIGNAL_CHANGED

        /* Latitude. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Latitude:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -90.0f, 90.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_gps_latitude_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 70, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set latitude center of the scenery (in degrees)"
        );
        gtk_widget_show(w);
	SET_EDITABLE_SIGNAL_CHANGED

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* Planetary radius. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Planetary Radius:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            100.0f, 1000.0f, 0.0f
        );
        /* Spin. */
        e->hdr_planetary_radius_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the radius of the planet (in meters), default value is 6371000"
        );
	gtk_widget_show(w);
	SET_EDITABLE_SIGNAL_CHANGED

        /* Cant. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Cant:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -360.0f, 360.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_cant_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 70, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the cant deviation angle (in degrees)"
        );
        gtk_widget_show(w);
	SET_EDITABLE_SIGNAL_CHANGED

        /* Elevation. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("MSL Elevation:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -1000000.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_elevation_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the mean sea level elevation of the scenery (in feet)"
        );
        gtk_widget_show(w);
	SET_EDITABLE_SIGNAL_CHANGED




        /* Textures frame. */
        w = gtk_frame_new("Textures");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;
	w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	w = gtk_label_new(
"These textures are loaded when the scene is loaded.  These textures\
 will be available to all objects in this\nscene.  The texture for\
 the ground tile should be specified here"
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	/* Editable clist. */
	heading[0] = "Name";
	heading[1] = "File";
	heading[2] = "Priority";
	e->hdr_textures_eclist = eclist = EditCListNew(
	    parent3,
	    -1, 150,
	    FALSE, FALSE, 0,
	    heading, 3,
	    EditorHdrEditCListChangedCB, e
	);
	EditCListMap(eclist);
	w = EDITCLIST_CLIST(eclist);
	if(w != NULL)
	{
	    clist = GTK_CLIST(w);
            gtk_clist_column_titles_show(clist);
            gtk_clist_column_titles_passive(clist);
            gtk_clist_set_row_height(clist, SE_LIST_ROW_SPACING);
            gtk_clist_set_reorderable(clist, FALSE);
	    gtk_clist_set_column_width(clist, 0, 170);
            gtk_clist_set_column_width(clist, 1, 250);
            gtk_clist_set_column_width(clist, 2, 70);
	}

	/* Registered locations. */
        w = gtk_frame_new("Registered Locations");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        w = gtk_label_new(
"These are locations used as starting/restarting positions.  The X\
 and Y coordinate values are in meters, the Z\ncoordinate value is\
 in feet, and the angles are in degrees"
        );
        gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        /* Editable clist. */
        heading[0] = "Name";
        heading[1] = "X";
        heading[2] = "Y";
        heading[3] = "Z";
        heading[4] = "Heading";
        heading[5] = "Pitch";
        heading[6] = "Bank";
        e->hdr_registered_locations_eclist = eclist = EditCListNew(
            parent3,
            -1, 150,
            FALSE, FALSE, 0,
            heading, 7,
            EditorHdrEditCListChangedCB, e
        );
        EditCListMap(eclist);
        w = EDITCLIST_CLIST(eclist);
        if(w != NULL)
        {
            clist = GTK_CLIST(w);
	    gtk_clist_column_titles_show(clist);
	    gtk_clist_column_titles_passive(clist);
	    gtk_clist_set_row_height(clist, SE_LIST_ROW_SPACING);
	    gtk_clist_set_reorderable(clist, FALSE);
            gtk_clist_set_column_width(clist, 0, 150);
            gtk_clist_set_column_width(clist, 1, 70);
            gtk_clist_set_column_width(clist, 2, 70);
            gtk_clist_set_column_width(clist, 3, 70);
            gtk_clist_set_column_width(clist, 4, 50);
            gtk_clist_set_column_width(clist, 5, 50);
            gtk_clist_set_column_width(clist, 6, 50);
        }

        /* Ground base frame. */
        w = gtk_frame_new("Ground Base");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* Ground is wather toggle. */
	e->hdr_ground_is_water_tb = w = gtk_check_button_new_with_label(
	    "Is Water"
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w,
"Check this to specify that the ground base is treated as water,\
 making the landing or driving on the ground base a \"splash\""
	);
        gtk_widget_show(w);

	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

        /* Tile Width. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Tile X Length:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_ground_tile_width_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
        gtk_widget_set_usize(w, 100, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the tile X length (in meters)"
        );
        gtk_widget_show(w);
        SET_EDITABLE_SIGNAL_CHANGED

        /* Tile Height. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Tile Y Length:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_ground_tile_height_spin = w = gtk_spin_button_new(
            adj, 1.0, 0
        );
        gtk_widget_set_usize(w, 100, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the tile Y length (in meters)"
        );
        gtk_widget_show(w);
        SET_EDITABLE_SIGNAL_CHANGED

        /* Tile close range. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Textured Tile Range:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_ground_tile_close_range_spin = w = gtk_spin_button_new(
            adj, 1.0, 2
        );
        gtk_widget_set_usize(w, 100, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the distance at which textured tile will be displayed (in meters)"
        );
        gtk_widget_show(w);
        SET_EDITABLE_SIGNAL_CHANGED

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* Ground color. */
	e->hdr_ground_color_btn = w = CSDColorButtonNew(
	    "Color:", -1,
	    e, EditorHdrGroundColorCB
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Texture name. */
        w = GUIPromptBar(
            NULL, "Texture Name:", NULL, &entry_rtn
	);
	e->hdr_ground_texture_entry = (GtkWidget *)entry_rtn;
	entry = (GtkEntry *)entry_rtn;
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	w = (GtkWidget *)entry_rtn;
	SET_EDITABLE_SIGNAL_CHANGED


        /* Map. */
        w = gtk_frame_new("Map");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

	/* Map display vbox. */
        w = gtk_vbox_new(TRUE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, FALSE, 0);
        gtk_widget_show(w);
	parent5 = w;

        e->hdr_map_da = w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, 80, 80);
	EditorSetDrawingAreaEvents(
	    w, GTK_SIGNAL_FUNC(EditorHdrMapDrawingAreaEventCB), e
	);
	gtk_container_add(GTK_CONTAINER(parent5), w);
	gtk_widget_show(w);

        /* Map values vbox. */
        w = gtk_vbox_new(TRUE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Map width. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("X Length:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_map_width_spin = w = gtk_spin_button_new(
            adj, 1.0, 2
        );
        gtk_widget_set_usize(w, 100, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the map's X length (in meters)"
        );
        gtk_widget_show(w);
        SET_EDITABLE_SIGNAL_CHANGED

        /* Map height. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Y Length:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->hdr_map_height_spin = w = gtk_spin_button_new(
            adj, 1.0, 2
        );
        gtk_widget_set_usize(w, 100, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Set the map's y length (in meters)"
        );
        gtk_widget_show(w);
        SET_EDITABLE_SIGNAL_CHANGED



#undef SET_TOGGLE_BUTTON_SIGNAL_CHANGED
#undef SET_EDITABLE_SIGNAL_CHANGED
}

/*
 *	Called by EditorBuildSceneryPage() to build the object values
 *	widgets.
 */
static void EditorBuildObjectValues(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
)
{
	gint border_major = 5, border_minor = 2;
	gpointer browse_rtn, entry_rtn, combo_rtn;
	GList *glist;
	GtkAdjustment *adj;
	GtkWidget *w, *parent2, *parent3, *parent4, *parent5;
	GtkEntry *entry;
	GtkCombo *combo;


        w = gtk_vbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent2 = w;

        /* Identity frame. */
        w = gtk_frame_new("Identity");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
	parent3 = w;

	/* Name. */
	w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	parent4 = w;
        /* Pixmap. */
        e->obj_icon = w = (GtkWidget *)GUICreateMenuItemIcon(
            (guint8 **)icon_object_32x32_xpm
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	/* Entry. */
	e->obj_name_entry = w = gtk_entry_new();
	gtk_widget_set_usize(w, 200, -1);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w,
"Enter a unique name to identify this object (no spaces, tabs, or\
 punctuations allowed) or leave blank for no name"
	);
        gtk_widget_show(w);

	/* Type. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	/* Combo. */
        glist = NULL;
        glist = g_list_append(glist, "Static");
        glist = g_list_append(glist, "Ground");
        glist = g_list_append(glist, "Premodeled");
        glist = g_list_append(glist, "Runway");
        glist = g_list_append(glist, "Helipad");
        glist = g_list_append(glist, "Human");
        glist = g_list_append(glist, "Smoke");
        glist = g_list_append(glist, "Fire");
        w = GUIComboCreate(
            "Type:",
            (const gchar *)glist->data,
            glist,
            g_list_length(glist),
            &combo_rtn,
            e,
            NULL,
	    EditorObjTypeChangedCB
        );
        g_list_free(glist);
        e->obj_type_combo = (GtkWidget *)combo_rtn;
        combo = (GtkCombo *)combo_rtn;
        entry = GTK_ENTRY(combo->entry);
        gtk_entry_set_editable(entry, FALSE);
        gtk_widget_set_usize(GTK_WIDGET(entry), 90, -1);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* Model file. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	w = GUIPromptBarWithBrowse(
	    NULL, "Model File:", NULL, &entry_rtn, &browse_rtn,
	    e, EditorObjModelBrowseCB
	);
	e->obj_model_entry = (GtkWidget *)entry_rtn;
	entry = (GtkEntry *)entry_rtn;
        gtk_widget_set_usize(GTK_WIDGET(entry), 145, -1);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);


        /* Display frame. */
        w = gtk_frame_new("Display");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
	parent3 = w;

        /* Range. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Label. */
        w = gtk_label_new("Range:");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_range_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(w,
"Set the visible range radius (or 0 to use the default range)"
        );
        gtk_widget_show(w);

        /* Range Far. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Label. */
        w = gtk_label_new("Range Far:");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_range_far_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(w,
"Set the visible range radius for displaying \"far model\" (or\
 or 0 to use the default far range)"
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* No depth test. */
	e->obj_no_depth_test_tb = w = gtk_check_button_new_with_label(
	    "No Depth Test"
	);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w,
"Check this to have the object always displayed regardless of depth"
	);
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Polygon offset. */
        e->obj_polygon_offset_tb = w = gtk_check_button_new_with_label(
            "Polygon Offset"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(EditorObjPolygonOffsetToggledCB), e
	);
        GUISetWidgetTip(w,
"Check this to enable GL polygon offset when drawing this object"
        );
        gtk_widget_show(w);

        /* Polygon offset reverse. */
        e->obj_polygon_offset_reverse_tb = w = gtk_check_button_new_with_label(
            "Reverse"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Check this to enable reverse GL polygon offset when drawing this object"
        );
        gtk_widget_show(w);

        /* Polygon offset write depth. */
        e->obj_polygon_offset_write_depth_tb = w = gtk_check_button_new_with_label(
            "Write Depth"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Check this to enable the writing of depth with GL polygon offset"
        );
        gtk_widget_show(w);


        /* Size frame. */
        w = gtk_frame_new("Size");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
	parent3 = w;

        /* Shape. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent4 = w;
        /* Combo. */
        glist = NULL;
        glist = g_list_append(glist, "Spherical");
        glist = g_list_append(glist, "Cylendrical");
        glist = g_list_append(glist, "Rectangular");
        w = GUIComboCreate(
            "Shape:",
            (const gchar *)glist->data,
            glist,
            g_list_length(glist),
            &combo_rtn,
            e,
            NULL,
            EditorObjSizeShapeChangedCB
        );
        g_list_free(glist);
        e->obj_size_shape_combo = (GtkWidget *)combo_rtn;
        combo = (GtkCombo *)combo_rtn;
        entry = GTK_ENTRY(combo->entry);
        gtk_entry_set_editable(entry, FALSE);
        gtk_widget_set_usize(GTK_WIDGET(entry), 90, -1);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent4 = w;

        /* Radius. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Radius:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_size_radius_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent4 = w;

        /* X min and max. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("X Min:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_size_x_min_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);
        /* Label. */
        w = gtk_label_new("Max:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_size_x_max_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent4 = w;

        /* Y min and max. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Y Min:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_size_y_min_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);
        /* Label. */
        w = gtk_label_new("Max:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_size_y_max_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent4 = w;

        /* Z min and max. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Z Min:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_size_z_min_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);
        /* Label. */
        w = gtk_label_new("Max:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_size_z_max_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);




	/* ********************************************************* */
        w = gtk_vbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent2 = w;

        /* Position frame. */
        w = gtk_frame_new("Position");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
	parent3 = w;

	/* Position row. */
	w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
	parent4 = w;

        /* X position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("X:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_x_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Y position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Y:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_y_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Z position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Z:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_z_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

	/* Direction row. */
	w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
	parent4 = w;

        /* Heading. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Heading:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 360.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_heading_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 65, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Pitch. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Pitch:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 360.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_pitch_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 65, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Bank. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Bank:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 360.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_bank_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 65, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        gtk_widget_show(w);


        /* Reference frame. */
        w = gtk_frame_new("Reference Object");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

        /* Hbox for reference object. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Reference object name label. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Name:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        /* Entry. */
        e->obj_ref_name_entry = w = gtk_entry_new();
        gtk_widget_set_usize(w, 200, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_entry_set_editable(GTK_ENTRY(w), FALSE);
        gtk_widget_show(w);
        /* Button. */
        e->obj_ref_change_btn = w = gtk_button_new_with_label(
            "Change"
        );
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(EditorObjReferenceChangeCB), e
        );
        gtk_widget_show(w);

        /* Reference object offset position row. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent4 = w;

        /* X offset position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("X:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_ref_x_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjReferenceAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Y offset position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Y:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_ref_y_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjReferenceAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Z offset position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Z:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->obj_ref_z_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjReferenceAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Offset direction row. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent4 = w;

        /* Offset heading. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Heading:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 360.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_ref_heading_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 65, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjReferenceAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Offset pitch. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Pitch:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 360.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_ref_pitch_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 65, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjReferenceAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Offset bank. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Bank:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 360.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_ref_bank_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 65, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjReferenceAdjValueChangedCB), e
        );
        gtk_widget_show(w);


        /* Helipad values frame. */
        e->obj_helipad_frame = w = gtk_frame_new("Helipad Values");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Type combo. */
        glist = NULL;
	glist = g_list_append(glist, "Ground Paved");
        glist = g_list_append(glist, "Ground Bare");
        glist = g_list_append(glist, "Building");
        glist = g_list_append(glist, "Vehicle");
        w = GUIComboCreate(
            "Style:",
            (const gchar *)glist->data,
            glist,
            g_list_length(glist),
            &combo_rtn,
            e,
            NULL,
            NULL
        );
        g_list_free(glist);
        e->obj_helipad_type_combo = (GtkWidget *)combo_rtn;
        combo = (GtkCombo *)combo_rtn;
        entry = GTK_ENTRY(combo->entry);
        gtk_entry_set_editable(entry, TRUE);
        gtk_widget_set_usize(GTK_WIDGET(entry), 250, -1);
        GUISetWidgetTip(
            GTK_WIDGET(entry),
"Select the helipad type (internally defined by Search and Rescue)"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* Edge lighting. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        e->obj_helipad_edge_lighting_tb = w = gtk_check_button_new_with_label(
            "Edge Lighting"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to display edge lighting around this helipad"
        );
        gtk_widget_show(w);

        /* Fuel, Repair, and Dropoff. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        e->obj_helipad_fuel_tb = w = gtk_check_button_new_with_label(
            "Fuel"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to allow helicopters to be refueled at this helipad"
        );
        gtk_widget_show(w);
        e->obj_helipad_repair_tb = w = gtk_check_button_new_with_label(
            "Repair"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to allow helicopters to be repaired at this helipad"
        );
        gtk_widget_show(w);
        e->obj_helipad_dropoff_tb = w = gtk_check_button_new_with_label(
            "Dropoff"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to allow the dropping off of passengers at this helipad"
        );
        gtk_widget_show(w);

        /* Restart point. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        e->obj_helipad_restart_point_tb = w = gtk_check_button_new_with_label(
            "Restart Point"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to mark the helipad as a restarting point, respawned\
 helicopters can restart at this helipad if it is nearby"
        );
        gtk_widget_show(w);

	/* Length and width. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Length. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Length:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_helipad_length_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Set the length of the helipad"
        );
        gtk_widget_show(w);
        /* Width. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Width:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_helipad_width_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Set the width of the helipad"
        );
        gtk_widget_show(w);

        /* Recession. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Recession. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Recession:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_helipad_recession_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
	    w,
"Set the amount of (positive) recession downward into the base (not\
 available on all helipad types)"
	);
        gtk_widget_show(w);

        /* Label. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        w = GUIPromptBar(NULL, "Label:", NULL, &entry_rtn);
        entry = (GtkEntry *)entry_rtn;
        e->obj_helipad_label_entry = (GtkWidget *)entry_rtn;
        gtk_widget_set_usize(GTK_WIDGET(entry), 250, -1);
        GUISetWidgetTip(
            GTK_WIDGET(entry),
"Enter the helipad label (no spaces, tabs, or punctuations allowed)"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);


        /* Runway values frame. */
        e->obj_runway_frame = w = gtk_frame_new("Runway Values");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Surface combo. */
        glist = NULL;
        glist = g_list_append(glist, "Paved");
        glist = g_list_append(glist, "Gravel");
        glist = g_list_append(glist, "Concrete");
        glist = g_list_append(glist, "Groved");
        w = GUIComboCreate(
            "Surface:",
            (const gchar *)glist->data,
            glist,
            g_list_length(glist),
            &combo_rtn,
            e,
            NULL,
            NULL
        );
        g_list_free(glist);
        e->obj_runway_surface_combo = (GtkWidget *)combo_rtn;
        combo = (GtkCombo *)combo_rtn;
        entry = GTK_ENTRY(combo->entry);
        gtk_entry_set_editable(entry, FALSE);
        gtk_widget_set_usize(GTK_WIDGET(entry), 250, -1);
        GUISetWidgetTip(
            GTK_WIDGET(entry),
"Select the runway surface type"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Length. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Length:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_runway_length_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(
	    w,
"Set the length of the runway"
        );
        gtk_widget_show(w);
        /* Width. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Width. */
        w = gtk_label_new("Width:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_runway_width_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(
            w,
"Set the width of the runway"
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Dashes. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Dashes:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_runway_dashes_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 50, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(
            w,
"Set the number of dashes (or 0 for no dashes)"
        );
        gtk_widget_show(w);

        /* Edge light spacing. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Edge Light Spacing:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000.0f,
            1.0f, 10.0f, 0.0f
        );
        /* Spin. */
        e->obj_runway_edge_light_spacing_spin = w = gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(
            w,
"Set the edge light spacing (or 0 for no edge lighting)"
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* Thresholds toggle. */
        e->obj_runway_thresholds_tb = w = gtk_check_button_new_with_label(
            "Thresholds"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
	    w,
"Check this to display thesholds (not displaced thresholds) at the ends\
 of the runway"
	);
        gtk_widget_show(w);

        /* Borders toggle. */
        e->obj_runway_borders_tb = w = gtk_check_button_new_with_label(
            "Borders"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to display borders along the sides of the runway"
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Touch down markers toggle. */
        e->obj_runway_td_markers_tb = w = gtk_check_button_new_with_label(
            "Touch Down Markers"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to display touch down markers on the runway"
        );
        gtk_widget_show(w);

        /* Midway markers toggle. */
        e->obj_runway_midway_markers_tb = w = gtk_check_button_new_with_label(
            "Midway Markers"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to display midway markers on the runway"
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* North glideslope toggle. */
        e->obj_runway_north_gs_tb = w = gtk_check_button_new_with_label(
            "North Glide Slope"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to display the glide slope on the north end of the runway"
        );
        gtk_widget_show(w);

        /* South glideslope toggle. */
        e->obj_runway_south_gs_tb = w = gtk_check_button_new_with_label(
            "South Glide Slope"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to display the glide slope on the south end of the runway"
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* North label. */
        w = GUIPromptBar(NULL, "North Label:", NULL, &entry_rtn);
        entry = (GtkEntry *)entry_rtn;
        e->obj_runway_north_label_entry = (GtkWidget *)entry_rtn;
        gtk_widget_set_usize(GTK_WIDGET(entry), 100, -1);
        GUISetWidgetTip(
            GTK_WIDGET(entry),
"Enter the label for the \"north\" end of the runway\
 (no spaces, tabs, or punctuations allowed)"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* South label. */
        w = GUIPromptBar(NULL, "South Label:", NULL, &entry_rtn);
        entry = (GtkEntry *)entry_rtn;
        e->obj_runway_south_label_entry = (GtkWidget *)entry_rtn;
        gtk_widget_set_usize(GTK_WIDGET(entry), 100, -1);
        GUISetWidgetTip(
            GTK_WIDGET(entry),
"Enter the label for the \"south\" end of the runway\
 (no spaces, tabs, or punctuations allowed)"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* North displaced threshold spin. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        w = gtk_label_new("Displaced Threshold North:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000.0f,
            1.0f, 10.0f, 0.0f
        );
        e->obj_runway_north_displaced_threshold_spin = w =
	    gtk_spin_button_new(adj, 1.0, 2);
	gtk_widget_set_usize(w, 70, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(
            w,
"Set displaced threshold for the north end of the runway (set to 0\
 for no displaced threshold)"
        );
        gtk_widget_show(w);

        /* South displaced threshold spin. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        w = gtk_label_new("South:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000.0f,
            1.0f, 10.0f, 0.0f
        );
        e->obj_runway_south_displaced_threshold_spin = w =
            gtk_spin_button_new(adj, 1.0, 2);
        gtk_widget_set_usize(w, 70, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorObjAdjValueChangedCB), e
        );
        GUISetWidgetTip(
            w,
"Set displaced threshold for the south end of the runway (set to 0\
 for no displaced threshold)"
        );
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* North end approach lighting. */
	w = gtk_label_new("North Approach Lighting:");
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

        /* End toggle. */
        e->obj_runway_north_lighting_end_tb = w =
	    gtk_check_button_new_with_label("End");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Check this to add end lights to the runway"
        );
        gtk_widget_show(w);

        /* Tracer toggle. */
        e->obj_runway_north_lighting_tracer_tb = w =
            gtk_check_button_new_with_label("Tracer");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to add tracer lights to the runway"
        );
        gtk_widget_show(w);

        /* Align toggle. */
        e->obj_runway_north_lighting_align_tb = w =
            gtk_check_button_new_with_label("Align");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to add alignment lighting to the runway"
        );
        gtk_widget_show(w);

        /* ILS glide toggle. */
        e->obj_runway_north_lighting_ils_glide_tb = w =
            gtk_check_button_new_with_label("ILS Glide");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to add ILS glide lighting to the runway"
        );
        gtk_widget_show(w);


        /* South end approach lighting. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        w = gtk_label_new("South Approach Lighting:");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* End toggle. */
        e->obj_runway_south_lighting_end_tb = w =
            gtk_check_button_new_with_label("End");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(w,
"Check this to add end lights to the runway"
        );
        gtk_widget_show(w);

        /* Tracer toggle. */
        e->obj_runway_south_lighting_tracer_tb = w =
            gtk_check_button_new_with_label("Tracer");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to add tracer lights to the runway"
        );
        gtk_widget_show(w);

        /* Align toggle. */
        e->obj_runway_south_lighting_align_tb = w =
            gtk_check_button_new_with_label("Align");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to add alignment lighting to the runway"
        );
        gtk_widget_show(w);

        /* ILS glide toggle. */
        e->obj_runway_south_lighting_ils_glide_tb = w =
            gtk_check_button_new_with_label("ILS Glide");
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        GUISetWidgetTip(
            w,
"Check this to add ILS glide lighting to the runway"
        );
        gtk_widget_show(w);


        /* Human values frame. */
	e->obj_human_frame = w = gtk_frame_new("Human Values");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Type combo. */
        glist = EditorFIOHumanGetTypes(NULL);
        w = GUIComboCreate(
            "Type:",
            (const gchar *)glist->data,
            glist,
            g_list_length(glist),
            &combo_rtn,
            e,
            NULL,
	    NULL
        );
	g_list_foreach(glist, (GFunc)g_free, NULL);
        g_list_free(glist);
        e->obj_human_type_combo = (GtkWidget *)combo_rtn;
        combo = (GtkCombo *)combo_rtn;
        entry = GTK_ENTRY(combo->entry);
        gtk_entry_set_editable(entry, TRUE);
        gtk_widget_set_usize(GTK_WIDGET(entry), 250, -1);
	GUISetWidgetTip(
	    GTK_WIDGET(entry),
"Select the preset human type (defined in human.ini)"
	);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

	/* Need rescue. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	e->obj_human_need_rescue_tb = w = gtk_check_button_new_with_label(
	    "Need Rescue"
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

	/* Sit, sit up, sit down, and lying. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        e->obj_human_sit_tb = w = gtk_check_button_new_with_label(
            "Sit"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        e->obj_human_sit_up_tb = w = gtk_check_button_new_with_label(
            "Sit Up"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        e->obj_human_sit_down_tb = w = gtk_check_button_new_with_label(
            "Sit Down"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        e->obj_human_lying_tb = w = gtk_check_button_new_with_label(
            "Lying"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* Alert, aware, and in water. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        e->obj_human_alert_tb = w = gtk_check_button_new_with_label(
            "Alert"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        e->obj_human_aware_tb = w = gtk_check_button_new_with_label(
            "Aware"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        e->obj_human_in_water_tb = w = gtk_check_button_new_with_label(
            "In Water"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* On streatcher. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        e->obj_human_on_streatcher_tb = w = gtk_check_button_new_with_label(
            "On Streatcher"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* Run, run towards, and run away. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
#if 0
        e->obj_human_run_tb = w = gtk_check_button_new_with_label(
            "Run"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
#endif
        e->obj_human_run_towards_tb = w = gtk_check_button_new_with_label(
            "Run Towards"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        e->obj_human_run_away_tb = w = gtk_check_button_new_with_label(
            "Run Away"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* Message enter. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	parent4 = w;
	w = GUIPromptBar(NULL, "Message Enter:", NULL, &entry_rtn);
	entry = (GtkEntry *)entry_rtn;
	e->obj_human_message_enter_entry = (GtkWidget *)entry_rtn;
	gtk_widget_set_usize(GTK_WIDGET(entry), 250, -1);
	GUISetWidgetTip(
	    GTK_WIDGET(entry),
"Enter the message that will be shown when the human enters the\
 aircraft (spaces and tabs are allowed), or leave blank for no message"
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


        /* Preset model values frame. */
        e->obj_preset_model_frame = w = gtk_frame_new("Preset Model");
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
        gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
        parent3 = w;
        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_container_add(GTK_CONTAINER(parent3), w);
        gtk_widget_show(w);
        parent3 = w;

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* Preset model types combo. */
	{
	    gint i;
	    gchar *names[] = SE_SAR_PRESET_MODEL_NAMES;
	    glist = NULL;
	    for(i = 0; names[i] != NULL; i++)
		glist = g_list_append(glist, STRDUP(names[i]));
	}	    
        w = GUIComboCreate(
            "Type:",
            (const gchar *)glist->data,
            glist,
            g_list_length(glist),
            &combo_rtn,
            e,
	    NULL,
            EditorObjPresetModelTypeChangedCB
        );
        g_list_foreach(glist, (GFunc)g_free, NULL);
        g_list_free(glist);
        e->obj_preset_model_type_combo = (GtkWidget *)combo_rtn;
        combo = (GtkCombo *)combo_rtn;
        entry = GTK_ENTRY(combo->entry);
        gtk_entry_set_editable(entry, TRUE);
        gtk_widget_set_usize(GTK_WIDGET(entry), 250, -1);
        GUISetWidgetTip(
            GTK_WIDGET(entry),
"Select the preset model type"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	/* Preset model values entry. */
        w = GUIPromptBar(NULL, "Values:", NULL, &entry_rtn);
        entry = (GtkEntry *)entry_rtn;
        e->obj_preset_model_values_entry = (GtkWidget *)entry_rtn;
        gtk_widget_set_usize(GTK_WIDGET(entry), 250, -1);
        GUISetWidgetTip(
            GTK_WIDGET(entry),
"Enter a space separated list of values for the preset model"
        );
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);


}

/*
 *	Called by EditorNew() to build the scenery page.
 */
void EditorBuildSceneryPage(
        editor_struct *e, GtkWidget *parent, GdkWindow *window,
        const options_struct *opt
)
{
	gint border_major = 5, border_minor = 2;
	gpointer combo_rtn;
	GList *glist, *glist2;
	GtkAdjustment *adj;
	GtkWidget *w, *menu, *parent2, *parent3, *parent4, *parent5;
	GtkEntry *entry;
	GtkCombo *combo;
	GtkCList *clist;

	/* Vpaned to separate view and list with the object values
	 * widgets.
	 */
	e->scenery_vpaned = w = gtk_vpaned_new();
        gtk_paned_set_handle_size(
            GTK_PANED(w), SE_DEF_PANED_HANDLE_SIZE
        );
        gtk_paned_set_gutter_size(
            GTK_PANED(w), SE_DEF_PANED_GUTTER_SIZE
        );
        gtk_paned_set_position(
	    GTK_PANED(w),
	    (opt != NULL) ? opt->editor_last_scenery_paned_position : 330
	);
        gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent = w;


	/* Horizontal paned widget to separate objects list and view. */
	e->view_hpaned = w = gtk_hpaned_new();
	gtk_paned_set_handle_size(
	    GTK_PANED(w), SE_DEF_PANED_HANDLE_SIZE
	);
        gtk_paned_set_gutter_size(
            GTK_PANED(w), SE_DEF_PANED_GUTTER_SIZE
        );
        gtk_paned_set_position(
            GTK_PANED(w),
            (opt != NULL) ? opt->editor_last_view_paned_position : 180
        );
        gtk_paned_add1(GTK_PANED(parent), w);
        gtk_widget_show(w);
	parent2 = w;

	/* Scrolled window for objects clist vbox. */
        w = gtk_scrolled_window_new(NULL, NULL);
        gtk_scrolled_window_set_policy(
            GTK_SCROLLED_WINDOW(w),
            GTK_POLICY_AUTOMATIC,
            GTK_POLICY_AUTOMATIC
        );
        gtk_paned_add1(GTK_PANED(parent2), w);
        gtk_widget_show(w);
        parent3 = w;

	/* Objects clist. */
	e->obj_clist = w = gtk_clist_new(1);
	clist = GTK_CLIST(w);
        gtk_clist_set_row_height(clist, SE_LIST_ROW_SPACING);
	gtk_clist_set_column_width(clist, 0, 10);
        gtk_widget_add_events(
            w,
            GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
            GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_press_event",
            GTK_SIGNAL_FUNC(EditorObjListKeyEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(EditorObjListKeyEventCB), e
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(EditorObjListButtonEventCB), e
        );
        gtk_signal_connect_after(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(EditorObjListButtonEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "select_row",
            GTK_SIGNAL_FUNC(EditorCListSelectRowCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "unselect_row",
            GTK_SIGNAL_FUNC(EditorCListUnselectRowCB), e
        );
        gtk_clist_set_selection_mode(clist, GTK_SELECTION_EXTENDED);
        gtk_container_add(GTK_CONTAINER(parent3), w);
#if 0
        if(TRUE)
        {
            const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",		0,	YC_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",	0,	YC_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",		0,	YC_DND_TYPE_INFO_STRING}
                };
	    GUIDNDSetTar(
                w,
                dnd_tar_types,
                sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
                GDK_ACTION_COPY,	/* Actions. */
                GDK_ACTION_COPY,	/* Default action if same. */
                GDK_ACTION_COPY,	/* Default action. */
                EditorObjListDragDataReceivedCB,
                e
            );
            gtk_signal_connect_after(
                GTK_OBJECT(w), "drag_motion",
                GTK_SIGNAL_FUNC(EditorObjListDragMotionCB), e
            );
        }
#endif
        gtk_widget_show(w);



	/* Vbox for view widgets. */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_paned_add2(GTK_PANED(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Hbox for view position widgets. */
	w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

	/* Units. */
	e->units = (opt != NULL) ?
	    opt->editor_last_units : EDITOR_UNITS_METRIC;
        glist = NULL;
        glist = g_list_append(glist, "Metric");
        glist = g_list_append(glist, "Metric (Alt Feet)");
        glist = g_list_append(glist, "English");
        w = GUIComboCreate(
            "Units:",
            (const gchar *)glist->data,
            glist,
            g_list_length(glist),
            &combo_rtn,
            e,
            NULL,
            EditorViewUnitsChangedCB
        );
        e->view_units_combo = (GtkWidget *)combo_rtn;
        combo = (GtkCombo *)combo_rtn;
        entry = GTK_ENTRY(combo->entry);
        gtk_entry_set_editable(entry, FALSE);
	gtk_widget_set_usize(GTK_WIDGET(entry), 105, -1);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	glist2 = g_list_nth(glist, e->units);
	if(glist2 != NULL)
	    gtk_entry_set_text(entry, (const gchar *)glist2->data);
	g_list_free(glist);

	/* X position. */
        w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	parent5 = w;
	/* Label. */
	w = gtk_label_new("X:");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f, -10000000.0f, 10000000.0f,
	    10.0f, 100.0f, 0.0f
	);
	/* Spin. */
	e->view_x_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(EditorViewAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Y position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Y:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->view_y_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorViewAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Zoom. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Zoom:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            300.0f, 1.0f, 1000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->view_zoom_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 70, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorViewAdjValueChangedCB), e
        );
        gtk_widget_show(w);


        /* Hbox for view cursor widgets. */
        w = gtk_hbox_new(FALSE, border_major);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;

        /* Grid spacing. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Grid:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            100.0f, 1.0f, 1000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->view_grid_spacing_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 70, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorViewAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Cursor x position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Cursor X:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->view_cursor_x_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorViewAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Cursor y position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Y:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, -10000000.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->view_cursor_y_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorViewAdjValueChangedCB), e
        );
        gtk_widget_show(w);

        /* Cursor z position. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent5 = w;
        /* Label. */
        w = gtk_label_new("Z:");
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0f, 0.0f, 10000000.0f,
            10.0f, 100.0f, 0.0f
        );
        /* Spin. */
        e->view_cursor_z_spin = w = gtk_spin_button_new(adj, 1.0, 0);
        gtk_widget_set_usize(w, 90, -1);
        gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(EditorViewAdjValueChangedCB), e
        );
        gtk_widget_show(w);


	/* Frame for view. */
        w = gtk_frame_new(NULL);
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent4 = w;
	/* View. */
	e->view_da = w = gtk_drawing_area_new();
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
        gtk_widget_add_events(
            w,
            GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
            GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
            GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
            GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK |
            GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
            GDK_FOCUS_CHANGE_MASK
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "configure_event",
            GTK_SIGNAL_FUNC(EditorViewConfigureEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "expose_event",
            GTK_SIGNAL_FUNC(EditorViewExposeEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_press_event",
            GTK_SIGNAL_FUNC(EditorViewKeyEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(EditorViewKeyEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(EditorViewButtonEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(EditorViewButtonEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "motion_notify_event",
            GTK_SIGNAL_FUNC(EditorViewMotionEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "focus_in_event",
            GTK_SIGNAL_FUNC(EditorViewFocusEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "focus_out_event",
            GTK_SIGNAL_FUNC(EditorViewFocusEventCB), e
        );
        gtk_container_add(GTK_CONTAINER(parent4), w);
        if(TRUE)
        {
            const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",		0,	SE_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",	0,	SE_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",		0,	SE_DND_TYPE_INFO_STRING}
	    };
	    GUIDNDSetTar(
                w,
                dnd_tar_types,
                sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
                GDK_ACTION_COPY,            /* Actions. */
                GDK_ACTION_COPY,            /* Default action if same. */
                GDK_ACTION_COPY,            /* Default action. */
                EditorViewDragDataReceivedCB,
		e
	    );
            gtk_signal_connect_after(
                GTK_OBJECT(w), "drag_motion",
                GTK_SIGNAL_FUNC(EditorViewDragMotionCB), e
	    );
	}
        gtk_widget_show(w);
	e->view_pm = NULL;


        /* Scrolled window for object value widgets. */
        e->obj_sw = w = gtk_scrolled_window_new(NULL, NULL);
        gtk_scrolled_window_set_policy(
            GTK_SCROLLED_WINDOW(w),
            GTK_POLICY_AUTOMATIC,
            GTK_POLICY_AUTOMATIC
        );
        gtk_paned_add2(GTK_PANED(parent), w);
        gtk_widget_show(w);
        parent2 = w;
	/* Event box. */
	e->obj_sw_eb = w = gtk_event_box_new();
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
	gtk_widget_add_events(
	    w,
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
	);
        gtk_signal_connect(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "motion_notify_event",
            GTK_SIGNAL_FUNC(EditorScrolledWindowEventCB), e
        );
        gtk_scrolled_window_add_with_viewport(
            GTK_SCROLLED_WINDOW(parent2), w
        );
        gtk_widget_show(w);
        parent2 = w;
        /* Hbox for object widgets. */
        e->obj_main_box = w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
        gtk_widget_show(w);
        parent2 = w;

	/* Build object values widgets. */
	EditorBuildObjectValues(e, parent2, window, opt);


	/* Build object right click menu. */
        e->obj_menu = menu = GUIMenuCreate();
        if(menu != NULL)
        {
	    guint8 **icon;
	    const gchar *label;
	    gint accel_key = 0;
	    guint accel_mods = 0;
	    GtkWidget *fw;
	    void (*func_cb)(GtkWidget *, gpointer);
	    gpointer mclient_data = e;
	    GtkAccelGroup *accelgrp = NULL;

#define DO_ADD_MENU_ITEM_LABEL  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp, \
  icon, label, accel_key, accel_mods, (gpointer *)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_ITEM_CHECK  \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp, \
  icon, label, accel_key, accel_mods, (gpointer *)&fw, \
  mclient_data, func_cb \
 ); }
#define DO_ADD_MENU_SEP         \
{ w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); }
            icon = (guint8 **)icon_cut_20x20_xpm;
            label = "Cut";
            func_cb = EditorCutCB;
            DO_ADD_MENU_ITEM_LABEL
            e->obj_cut_mi = w;

            icon = (guint8 **)icon_copy_20x20_xpm;
            label = "Copy";
            func_cb = EditorCopyCB;
            DO_ADD_MENU_ITEM_LABEL
            e->obj_copy_mi = w;

            icon = (guint8 **)icon_paste_20x20_xpm;
            label = "Paste";
            func_cb = EditorPasteCB;
            DO_ADD_MENU_ITEM_LABEL
            e->obj_paste_mi = w;

	    DO_ADD_MENU_SEP

            icon = (guint8 **)icon_object_20x20_xpm;
            label = "New Object";
            func_cb = EditorNewObjectCB;
            DO_ADD_MENU_ITEM_LABEL
            e->obj_new_object_mi = w;

            icon = (guint8 **)icon_goto_20x20_xpm;
            label = "GoTo  Object";
            func_cb = EditorGoToSelectedCB;
            DO_ADD_MENU_ITEM_LABEL
            e->obj_goto_selected_object_mi = w;

            icon = NULL;
            label = "Move Object To Cursor";
            func_cb = EditorMoveToCursorCB;
            DO_ADD_MENU_ITEM_LABEL
            e->obj_move_object_to_cursor_mi = w;

            icon = (guint8 **)icon_cancel_20x20_xpm;
            label = "Delete Object";
            func_cb = EditorDeleteObjectCB;
            DO_ADD_MENU_ITEM_LABEL
            e->obj_delete_object_mi = w;

	    DO_ADD_MENU_SEP

	    icon = NULL;
	    label = "Show Map";
	    func_cb = EditorViewDisplayMapCB;
	    DO_ADD_MENU_ITEM_CHECK
	    e->obj_view_map_micheck = w;

            icon = NULL;
            label = "Show Object Labels";
            func_cb = EditorViewDisplayObjectLabelsCB;
            DO_ADD_MENU_ITEM_CHECK
            e->obj_view_object_labels_micheck = w;

            icon = NULL;
            label = "Show Position Label";
            func_cb = EditorViewDisplayPositionLabelCB;
            DO_ADD_MENU_ITEM_CHECK
            e->obj_view_position_label_micheck = w;

            icon = NULL;
            label = "Show Registered Locations";
            func_cb = EditorViewDisplayRegisteredLocationsCB;
            DO_ADD_MENU_ITEM_CHECK
            e->obj_view_registered_locations_micheck = w;

#undef DO_ADD_MENU_SEP
#undef DO_ADD_MENU_ITEM_CHECK
#undef DO_ADD_MENU_ITEM_LABEL
	}
}

/*
 *	Create a new editor.
 */
editor_struct *EditorNew(gpointer core_ptr)
{
	gint border_major = 5;
        GdkWindow *window;
        GtkAccelGroup *accelgrp;
	GtkStyle *style;
        GtkWidget *w, *parent, *parent2, *parent3;
	GtkWidget *main_vbox, *toplevel;
	GtkNotebook *notebook;
	options_struct *opt;
        se_core_struct *cp = SE_CORE(core_ptr);
        editor_struct *e = EDITOR(g_malloc0(sizeof(editor_struct)));
        if(e == NULL)
            return(NULL);

	opt = (cp != NULL) ? cp->options : NULL;

        /* Reset values. */
	e->map_state = FALSE;
	e->busy_count = 0;
	e->accelgrp = accelgrp = gtk_accel_group_new();
	e->core_ptr = core_ptr;
	e->processing = FALSE;
	e->page_num = 0;
	e->stop_count = 0;
        e->drag_button = 0;
        e->drag_key_state = 0x00000000;
        e->drag_last_x = 0;
        e->drag_last_y = 0;
	e->opid = NULL;
	e->total_opids = 0;

	e->hdr_map_buf = NULL;
	e->hdr_freeze_count = 0;

	e->view_buf = NULL;
	e->view_freeze_count = 0;
	e->view_display_flags = (opt != NULL) ? opt->editor_view_display_flags : 0;
	e->view_cursor_xw = 0;
	e->view_cursor_yw = 0;
	e->view_cursor_draw_toid = (guint)-1;
	e->view_scroll_toid = (guint)-1;
	e->view_rect_select = FALSE;
	e->view_rect_select_x0 = 0;
        e->view_rect_select_y0 = 0;
        e->view_rect_select_x1 = 0;
        e->view_rect_select_y1 = 0;

        e->obj_freeze_count = 0;
	e->obj_selected = -1;
	e->obj_sw_button = 0;
	e->obj_sw_button_last_x = 0;
	e->obj_sw_button_last_y = 0;

	e->units = EDITOR_UNITS_METRIC;
	e->has_changes = FALSE;
	e->scn_filename = NULL;
	e->scn_title = NULL;
	e->scn_description = NULL;



        /* Create toplevel. */
        e->toplevel = toplevel = w = gtk_window_new(GTK_WINDOW_DIALOG);
        gtk_widget_set_name(w, "scnedit window");
        if(opt != NULL)
        {
	    gtk_widget_set_usize(
		w, opt->editor_last_width, opt->editor_last_height
	    );
            gtk_widget_set_uposition(
                w, opt->editor_last_x, opt->editor_last_y
            );
        }
        gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);
        gtk_widget_realize(w);
        window = w->window;
        if(window != NULL)
        {
            gdk_window_set_decorations(
                window,
                GDK_DECOR_TITLE | GDK_DECOR_MENU | GDK_DECOR_MINIMIZE |
		GDK_DECOR_MAXIMIZE
            );
            gdk_window_set_functions(
                window,
                GDK_FUNC_MOVE | GDK_FUNC_RESIZE |
		GDK_FUNC_MINIMIZE | GDK_FUNC_MAXIMIZE | GDK_FUNC_CLOSE
            );
	    GUISetWMIcon(window, (guint8 **)scnedit_48x48_xpm);
        }
	style = gtk_widget_get_style(w);
        gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(EditorDeleteEventCB), e
        );
	parent = w;

	/* Load icons. */
#define LOAD_ICON(p,m,d)	\
{ if((window != NULL) && (style != NULL) && (*(d) != NULL)) { \
   *(p) = gdk_pixmap_create_from_xpm_d( \
    window, (m), &style->bg[GTK_STATE_NORMAL], (gchar **)(d) \
   ); \
} }
	LOAD_ICON(
	    &e->list_object_pixmap,
	    &e->list_object_mask,
	    icon_object_20x20_xpm
	);
        LOAD_ICON(
            &e->list_ground_pixmap,
            &e->list_ground_mask,
            icon_ground_20x20_xpm
        );
        LOAD_ICON(
            &e->list_helipad_pixmap,
            &e->list_helipad_mask,
            icon_helipad_20x20_xpm
        );
        LOAD_ICON(
            &e->list_runway_pixmap,
            &e->list_runway_mask,
            icon_runway_20x20_xpm
        );
        LOAD_ICON(
            &e->list_human_pixmap,
            &e->list_human_mask,
            icon_human_20x20_xpm
        );

#undef LOAD_ICON
#define LOAD_ICON_MASK(m,d)	\
{ if((window != NULL) && (style != NULL) && (*(d) != NULL)) { \
   GdkPixmap *pixmap = gdk_pixmap_create_from_xpm_d( \
    window, (m), &style->bg[GTK_STATE_NORMAL], (gchar **)(d) \
   ); \
   if(pixmap != NULL) gdk_pixmap_unref(pixmap); \
} }

	LOAD_ICON_MASK(
	    &e->view_object_mask,
	    icon_object_15x15_xpm
	);
        LOAD_ICON_MASK(
            &e->view_ground_mask,
            icon_ground_15x15_xpm
        );
        LOAD_ICON_MASK(
            &e->view_helipad_mask,
            icon_helipad_15x15_xpm
        );
        LOAD_ICON_MASK(
            &e->view_human_mask,
            icon_human_15x15_xpm
        );
        LOAD_ICON_MASK(
            &e->view_fire_mask,
            icon_fire_15x15_xpm
        );
        LOAD_ICON_MASK(
            &e->view_smoke_mask,
            icon_smoke_15x15_xpm
        );

#undef LOAD_ICON_MASK


	/* Main vbox. */
	e->main_vbox = main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
        gtk_widget_show(w);
	parent = w;

	/* Menu bar. */
	EditorBuildMenuBar(e, main_vbox, window, opt);

	/* Tool bar. */
	EditorBuildToolBar(e, main_vbox, window, opt);


	/* Frame for notebook. */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Notebook. */
	e->notebook = w = gtk_notebook_new();
	notebook = GTK_NOTEBOOK(w);
	gtk_notebook_set_tab_pos(notebook, GTK_POS_TOP);
        gtk_notebook_set_scrollable(notebook, TRUE);
        gtk_notebook_set_show_tabs(notebook, TRUE);
        gtk_notebook_set_show_border(notebook, TRUE);
        gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_signal_connect(
	    GTK_OBJECT(w), "switch_page",
	    GTK_SIGNAL_FUNC(EditorNotebookSwitchPageCB), e
	);
        gtk_widget_show(w);
	parent2 = w;

        /* Build header page. */
        w = gtk_label_new("Header");
        gtk_widget_show(w);
        parent3 = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(parent3), border_major);
        gtk_notebook_append_page(notebook, parent3, w);
        gtk_widget_show(parent3);
        EditorBuildHeaderPage(e, parent3, window, opt);

	/* Build scenery page. */
	w = gtk_label_new("Scenery");
	gtk_widget_show(w);
	parent3 = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(parent3), border_major);
        gtk_notebook_append_page(notebook, parent3, w);
        gtk_widget_show(parent3);
	EditorBuildSceneryPage(e, parent3, window, opt);



	/* Set main notebook page. */
	gtk_notebook_set_page(
	    GTK_NOTEBOOK(e->notebook),
	    (opt != NULL) ? opt->editor_last_page : 0
	);

	/* Status bar. */
	e->status_bar = StatusBarNew(main_vbox);
	StatusBarMap(e->status_bar);

	EditorStopReset(e, FALSE);
	EditorViewAdjValueChangedCB(NULL, e);
	EditorUpdateTitle(e);
	EditorUpdateMenus(e);

        if(e->view_cursor_draw_toid == (guint)-1)
	    e->view_cursor_draw_toid = gtk_timeout_add(
		SE_DEF_CURSOR_BLINK_RATE_MS,
		EditorViewCursorDrawTOCB,
		e
	    );

	return(e);
}

/*
 *      Updates the editor's title.
 */
void EditorUpdateTitle(editor_struct *e)
{
        gchar *title;
	const gchar *scn_title;
        GtkWidget *w = EDITOR_TOPLEVEL(e);
        if(w == NULL)
            return;

	/* Get scenery title. */
	if((e->scn_filename != NULL) ? (*e->scn_filename != '\0') : FALSE)
	{
	    scn_title = strrchr(e->scn_filename, DIR_DELIMINATOR);
	    if(scn_title != NULL)
		scn_title++;
	    else
		scn_title = e->scn_filename;
	}
	else
	{
	    scn_title = "Untitled";
	}

	/* Format and set title. */
        title = g_strdup_printf(
            "%s: %s%s",
            PROG_NAME, scn_title,
            EDITOR_HAS_CHANGES(e) ? " (*)" : ""
        );
        gtk_window_set_title(GTK_WINDOW(w), title);
        g_free(title);
}

/*
 *	Updates the editor's menus and widgets.
 */
void EditorUpdateMenus(editor_struct *e)
{
	gboolean active;
	gboolean sensitive;
	gulong flags;
	gint page, row, total_selected;
	GList *glist;
	GtkWidget *w;
	GtkCList *clist;
	GtkNotebook *notebook;
	toolbar_struct *tb;
	const editor_obj_struct *obj;

        if(e == NULL)
            return;

#define DO_SET_SENSITIVE        \
{ if(w != NULL) { gtk_widget_set_sensitive(w, sensitive); } }
#define DO_SET_MICHECK		\
{ if(w != NULL) { GTK_CHECK_MENU_ITEM(w)->active = active; } }

	notebook = (GtkNotebook *)EDITOR_NOTEBOOK(e);
	page = e->page_num;

	clist = (GtkCList *)e->obj_clist;
	glist = (clist != NULL) ? clist->selection_end : NULL;
	row = (glist != NULL) ? (gint)glist->data : -1;
	obj = EDITOR_OBJ(((clist != NULL) && (row > -1)) ?
	    gtk_clist_get_row_data(clist, row) : NULL
	);

	glist = (clist != NULL) ? clist->selection : NULL;
        total_selected = (glist != NULL) ? g_list_length(glist) : 0;

	tb = e->tool_bar;

	active = (page == 0) ? TRUE : FALSE;
	w = e->view_header_micheck;
	DO_SET_MICHECK
        active = (page == 1) ? TRUE : FALSE;
        w = e->view_scenery_micheck;
        DO_SET_MICHECK

	flags = e->view_display_flags;
	active = (flags & EDITOR_VIEW_DISPLAY_MAP) ? TRUE : FALSE;
        w = e->view_map_micheck;
        DO_SET_MICHECK
	w = e->obj_view_map_micheck;
	DO_SET_MICHECK
        active = (flags & EDITOR_VIEW_DISPLAY_OBJECT_LABELS) ? TRUE : FALSE;
        w = e->view_object_labels_micheck;
        DO_SET_MICHECK
        w = e->obj_view_object_labels_micheck;
        DO_SET_MICHECK
        active = (flags & EDITOR_VIEW_DISPLAY_POSITION_LABEL) ? TRUE : FALSE;
        w = e->view_position_label_micheck;
        DO_SET_MICHECK
        w = e->obj_view_position_label_micheck;
        DO_SET_MICHECK
        active = (flags & EDITOR_VIEW_DISPLAY_REGISTERED_LOCATIONS) ? TRUE : FALSE;
        w = e->view_registered_locations_micheck;
        DO_SET_MICHECK
        w = e->obj_view_registered_locations_micheck;
        DO_SET_MICHECK

	sensitive = (total_selected > 0) ? TRUE : FALSE;
	w = e->cut_mi;
	DO_SET_SENSITIVE
        w = e->obj_cut_mi;
        DO_SET_SENSITIVE
	ToolBarItemSetSensitiveID(
	    tb, EDITOR_OPID_CUT, sensitive
	);
        w = e->copy_mi;
        DO_SET_SENSITIVE
        w = e->obj_copy_mi;
        DO_SET_SENSITIVE
        ToolBarItemSetSensitiveID(
            tb, EDITOR_OPID_COPY, sensitive
        );
        w = e->goto_selected_object_mi;
        DO_SET_SENSITIVE
        w = e->obj_goto_selected_object_mi;
        DO_SET_SENSITIVE
        w = e->move_object_to_cursor_mi;
        DO_SET_SENSITIVE
        w = e->obj_move_object_to_cursor_mi;
        DO_SET_SENSITIVE
        w = e->delete_object_mi;
        DO_SET_SENSITIVE
        w = e->obj_delete_object_mi;
        DO_SET_SENSITIVE

	sensitive = (total_selected == 1) ? TRUE : FALSE;
        w = e->obj_icon;
        DO_SET_SENSITIVE
        w = e->obj_name_entry;
        DO_SET_SENSITIVE
        w = e->obj_type_combo;
        DO_SET_SENSITIVE
        w = e->obj_model_entry;
        DO_SET_SENSITIVE
        w = e->obj_x_spin;
        DO_SET_SENSITIVE
        w = e->obj_y_spin;
        DO_SET_SENSITIVE
        w = e->obj_z_spin;
        DO_SET_SENSITIVE
        w = e->obj_heading_spin;
        DO_SET_SENSITIVE
        w = e->obj_pitch_spin;
        DO_SET_SENSITIVE
        w = e->obj_bank_spin;
        DO_SET_SENSITIVE
        w = e->obj_range_spin;
        DO_SET_SENSITIVE
        w = e->obj_range_far_spin;
        DO_SET_SENSITIVE
        w = e->obj_no_depth_test_tb;
	DO_SET_SENSITIVE
        w = e->obj_polygon_offset_tb;
	DO_SET_SENSITIVE
/* Do not set sensitive for reverse and write depth. */
        w = e->obj_size_shape_combo;
        DO_SET_SENSITIVE
        w = e->obj_size_radius_spin;
        DO_SET_SENSITIVE
        w = e->obj_size_x_min_spin;
        DO_SET_SENSITIVE
        w = e->obj_size_x_max_spin;
        DO_SET_SENSITIVE
        w = e->obj_size_y_min_spin;
        DO_SET_SENSITIVE
        w = e->obj_size_y_max_spin;
        DO_SET_SENSITIVE
        w = e->obj_size_z_min_spin;
        DO_SET_SENSITIVE
        w = e->obj_size_z_max_spin;
        DO_SET_SENSITIVE
        w = e->obj_crash_type_combo;
        DO_SET_SENSITIVE
        w = e->obj_support_surface_tb;
        DO_SET_SENSITIVE
        w = e->obj_ref_name_entry;
        DO_SET_SENSITIVE
        w = e->obj_ref_change_btn;
        DO_SET_SENSITIVE
        w = e->obj_ref_x_spin;
        DO_SET_SENSITIVE
        w = e->obj_ref_y_spin;
        DO_SET_SENSITIVE
        w = e->obj_ref_z_spin;
        DO_SET_SENSITIVE
        w = e->obj_ref_heading_spin;
        DO_SET_SENSITIVE
        w = e->obj_ref_pitch_spin;
        DO_SET_SENSITIVE
        w = e->obj_ref_bank_spin;
        DO_SET_SENSITIVE

	/* Size shape dependent sensitivity. */
	if(total_selected == 1)
	{
	    switch(EditorGetSizeShape(e->obj_size_shape_combo))
	    {
	      case SAR_CONTACT_SHAPE_SPHERICAL:
		sensitive = TRUE;
		w = e->obj_size_radius_spin;
		DO_SET_SENSITIVE
                sensitive = FALSE;
                w = e->obj_size_x_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_x_max_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_y_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_y_max_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_z_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_z_max_spin;
                DO_SET_SENSITIVE
		w = e->obj_support_surface_tb;
                DO_SET_SENSITIVE
		break;

	      case SAR_CONTACT_SHAPE_CYLENDRICAL:
                sensitive = TRUE;
                w = e->obj_size_radius_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_z_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_z_max_spin;
                DO_SET_SENSITIVE
                w = e->obj_support_surface_tb;
                DO_SET_SENSITIVE
                sensitive = FALSE;
                w = e->obj_size_x_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_x_max_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_y_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_y_max_spin;
                DO_SET_SENSITIVE
                break;

	      case SAR_CONTACT_SHAPE_RECTANGULAR:
                sensitive = TRUE;
                w = e->obj_size_radius_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_x_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_x_max_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_y_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_y_max_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_z_min_spin;
                DO_SET_SENSITIVE
                w = e->obj_size_z_max_spin;
                DO_SET_SENSITIVE
                w = e->obj_support_surface_tb;
                DO_SET_SENSITIVE
                sensitive = FALSE;
                w = e->obj_size_radius_spin;
                DO_SET_SENSITIVE
                break;
	    }
	}

#undef DO_SET_SENSITIVE
}

/*
 *      Saves the editor's position to the configuration.
 */
void EditorSavePosition(editor_struct *e)
{
        GtkWidget *w;
	options_struct *opt;
        se_core_struct *core_ptr;


        if(e == NULL)
            return;

        core_ptr = SE_CORE(e->core_ptr);
        if(core_ptr == NULL)
            return;

        opt = (core_ptr != NULL) ? core_ptr->options : NULL;
        if(opt == NULL)
	    return;

	w = EDITOR_TOPLEVEL(e);
	if(w != NULL)
	{
	    GdkWindow *window = w->window;
            if(window != NULL)
                gdk_window_get_root_origin(
		    window, &opt->editor_last_x, &opt->editor_last_y
		);
            opt->editor_last_width = w->allocation.width;
            opt->editor_last_height = w->allocation.height;
	}

	w = e->notebook;
	if(w != NULL)
	    opt->editor_last_page = gtk_notebook_get_current_page(
		GTK_NOTEBOOK(w)
	    );

	w = e->scenery_vpaned;
	if(w != NULL)
	    opt->editor_last_scenery_paned_position =
		GTK_PANED(w)->child1_size;

        w = e->view_hpaned;
        if(w != NULL)
            opt->editor_last_view_paned_position =
                GTK_PANED(w)->child1_size;

	opt->editor_last_units = e->units;

	if(e->scn_filename != NULL)
	{
	    g_free(opt->last_scn_filename);
	    opt->last_scn_filename = STRDUP(e->scn_filename);
	}

	opt->editor_view_display_flags = e->view_display_flags;
	opt->editor_last_cursor_xw = e->view_cursor_xw;
	opt->editor_last_cursor_yw = e->view_cursor_yw;
	EditorViewGetViewPositionM(
	    e, &opt->editor_last_view_xm, &opt->editor_last_view_ym
	);
	EditorViewGetViewZoomM(e, &opt->editor_last_view_zoom);
	EditorViewGetViewGridSpacingM(e, &opt->editor_last_view_grid_spacing);
}

/*
 *	Sets the editor's busy state.
 */
void EditorSetBusy(editor_struct *e, gboolean is_busy)
{
        GdkCursor *cursor;
        GtkWidget *w;
        se_core_struct *core_ptr;


        if(e == NULL)
            return;

        core_ptr = SE_CORE(e->core_ptr);
        if(core_ptr == NULL)
            return;

        w = EDITOR_TOPLEVEL(e);
        if(w != NULL)
        {
            if(is_busy)
            {
                /* Increase busy count. */
                e->busy_count++;

                /* If already busy then don't change anything. */
                if(e->busy_count > 1)
                    return;

                cursor = core_ptr->cursor_busy;
            }
            else
            {
                /* Reduce busy count. */
                e->busy_count--;
                if(e->busy_count < 0)
                    e->busy_count = 0;

                /* If still busy do not change anything. */
                if(e->busy_count > 0)
                    return;

                cursor = NULL;  /* Use default cursor. */
            }

            /* Update toplevel window's cursor. */
            if(w->window != NULL)
            {
                gdk_window_set_cursor(w->window, cursor);
                gdk_flush();
            }
        }
}

/*
 *	Resets the stop_count on the editor and sets the stop function
 *	widgets sensitive or insensitive.
 */
void EditorStopReset(editor_struct *e, gboolean sensitive)
{
	toolbar_struct *tb;

	if(e == NULL)
	    return;

	e->stop_count = 0;

	tb = e->tool_bar;
        ToolBarItemSetSensitiveID(
            tb, EDITOR_OPID_STOP, sensitive
        );
}

/*
 *	Maps the editor.
 */
void EditorMap(editor_struct *e)
{
	GtkWidget *w = EDITOR_TOPLEVEL(e);
	if(w == NULL)
	    return;

	if(!EDITOR_IS_MAPPED(e))
	{
	    gtk_widget_show_raise(w);
	    e->map_state = TRUE;
	}
}

/*
 *	Unmaps the editor.
 */
void EditorUnmap(editor_struct *e)
{
        GtkWidget *w = EDITOR_TOPLEVEL(e);
        if(w == NULL)
            return;

        if(EDITOR_IS_MAPPED(e))
        {
            gtk_widget_hide(w);
            e->map_state = FALSE;
        }
}

/*
 *	Deletes the editor.
 */
void EditorDelete(editor_struct *e)
{
	gint i;
	GdkBitmap **bm;
	GdkPixmap **pm;
        GtkWidget **w;

#define DO_DESTROY_WIDGET	\
{ if(*w != NULL) { gtk_widget_destroy(*w); *w = NULL; } }
#define UNREF_PIXMAP_MASK	\
{ if(*pm != NULL) { gdk_pixmap_unref(*pm); *pm = NULL; } \
  if(*bm != NULL) { gdk_bitmap_unref(*bm); *bm = NULL; } \
}
#define UNREF_PIXMAP		\
{ if(*pm != NULL) { gdk_pixmap_unref(*pm); *pm = NULL; } }
#define UNREF_MASK		\
{ if(*bm != NULL) { gdk_bitmap_unref(*bm); *bm = NULL; } }


        if(e == NULL)
            return;

	/* Delete all objects and scenery values. */
	EditorSceneryUnload(e);

	REMOVE_TOID(&e->view_scroll_toid);
	REMOVE_TOID(&e->view_cursor_draw_toid);

	/* Menu bar menu items. */
	e->new_mi =
        e->open_mi =
        e->save_mi =
        e->save_as_mi =
        e->exit_mi =
        e->cut_mi =
        e->copy_mi =
        e->paste_mi =
        e->new_object_mi =
        e->goto_selected_object_mi =
        e->move_object_to_cursor_mi =
        e->delete_object_mi =
        e->view_header_micheck =
        e->view_scenery_micheck =
        e->view_map_micheck =
        e->view_object_labels_micheck =
        e->view_position_label_micheck =
        e->view_registered_locations_micheck =
	e->refresh_mi =
	e->check_scenery_mi =
	e->run_sar_mi = NULL;

	/* Delete status bar. */
	StatusBarDelete(e->status_bar);
	e->status_bar = NULL;

	w = &e->obj_menu;
	e->obj_cut_mi =
	e->obj_copy_mi =
        e->obj_paste_mi =
        e->obj_new_object_mi =
        e->obj_goto_selected_object_mi =
        e->obj_move_object_to_cursor_mi =
        e->obj_delete_object_mi =
	e->obj_view_map_micheck =
        e->obj_view_object_labels_micheck =
        e->obj_view_position_label_micheck =
        e->obj_view_registered_locations_micheck = NULL;
	DO_DESTROY_WIDGET

        w = &e->obj_helipad_type_combo;
	DO_DESTROY_WIDGET
        w = &e->obj_helipad_edge_lighting_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_fuel_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_repair_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_dropoff_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_restart_point_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_length_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_width_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_recession_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_label_entry;
        DO_DESTROY_WIDGET
        w = &e->obj_helipad_frame;
        DO_DESTROY_WIDGET

	w = &e->obj_runway_surface_combo;
	DO_DESTROY_WIDGET
        w = &e->obj_runway_length_spin;
	DO_DESTROY_WIDGET
        w = &e->obj_runway_width_spin;
        DO_DESTROY_WIDGET
	w = &e->obj_runway_dashes_spin;
	DO_DESTROY_WIDGET
	w = &e->obj_runway_edge_light_spacing_spin;
	DO_DESTROY_WIDGET
        w = &e->obj_runway_thresholds_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_borders_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_td_markers_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_midway_markers_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_north_gs_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_south_gs_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_north_label_entry;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_south_label_entry;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_north_displaced_threshold_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_south_displaced_threshold_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_north_lighting_end_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_north_lighting_tracer_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_north_lighting_align_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_north_lighting_ils_glide_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_south_lighting_end_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_south_lighting_tracer_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_south_lighting_align_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_south_lighting_ils_glide_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_runway_frame;
        DO_DESTROY_WIDGET

	w = &e->obj_human_type_combo;
        DO_DESTROY_WIDGET
        w = &e->obj_human_need_rescue_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_sit_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_sit_down_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_sit_up_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_lying_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_alert_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_aware_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_in_water_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_on_streatcher_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_run_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_run_towards_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_run_away_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_human_message_enter_entry;
	DO_DESTROY_WIDGET
        w = &e->obj_human_frame;
        DO_DESTROY_WIDGET

        w = &e->obj_preset_model_type_combo;
        DO_DESTROY_WIDGET
        w = &e->obj_preset_model_values_entry;
        DO_DESTROY_WIDGET
        w = &e->obj_preset_model_frame;
        DO_DESTROY_WIDGET

        w = &e->obj_icon;
        DO_DESTROY_WIDGET
        w = &e->obj_name_entry;
        DO_DESTROY_WIDGET
        w = &e->obj_type_combo;
        DO_DESTROY_WIDGET
        w = &e->obj_model_entry;
        DO_DESTROY_WIDGET
        w = &e->obj_x_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_y_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_z_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_heading_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_pitch_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_bank_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_range_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_range_far_spin;
        DO_DESTROY_WIDGET
	w = &e->obj_no_depth_test_tb;
	DO_DESTROY_WIDGET
        w = &e->obj_polygon_offset_tb;
	DO_DESTROY_WIDGET
        w = &e->obj_polygon_offset_reverse_tb;
	DO_DESTROY_WIDGET
        w = &e->obj_polygon_offset_write_depth_tb;
	DO_DESTROY_WIDGET
        w = &e->obj_size_shape_combo;
        DO_DESTROY_WIDGET
        w = &e->obj_size_radius_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_size_x_min_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_size_x_max_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_size_y_min_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_size_y_max_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_size_z_min_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_size_z_max_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_crash_type_combo;
        DO_DESTROY_WIDGET
        w = &e->obj_support_surface_tb;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_name_entry;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_change_btn;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_x_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_y_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_z_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_heading_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_pitch_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_ref_bank_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_main_box;
        DO_DESTROY_WIDGET
	w = &e->obj_sw_eb;
	DO_DESTROY_WIDGET
        w = &e->obj_sw;
        DO_DESTROY_WIDGET

        w = &e->view_da;
        DO_DESTROY_WIDGET
        w = &e->view_units_combo;
        DO_DESTROY_WIDGET
        w = &e->view_x_spin;
        DO_DESTROY_WIDGET
        w = &e->view_y_spin;
        DO_DESTROY_WIDGET
        w = &e->view_zoom_spin;
        DO_DESTROY_WIDGET
        w = &e->view_grid_spacing_spin;
        DO_DESTROY_WIDGET
        w = &e->view_cursor_x_spin;
        DO_DESTROY_WIDGET
        w = &e->view_cursor_y_spin;
        DO_DESTROY_WIDGET
        w = &e->view_cursor_z_spin;
        DO_DESTROY_WIDGET
        w = &e->obj_clist;
        DO_DESTROY_WIDGET
	w = &e->view_hpaned;
	DO_DESTROY_WIDGET
	w = &e->scenery_vpaned;
	DO_DESTROY_WIDGET


	EditCListDelete(e->hdr_registered_locations_eclist);
	e->hdr_registered_locations_eclist = NULL;
	EditCListDelete(e->hdr_textures_eclist);
	e->hdr_textures_eclist = NULL;
        w = &e->hdr_name_entry;
        DO_DESTROY_WIDGET
        w = &e->hdr_gps_longitude_spin;
        DO_DESTROY_WIDGET
        w = &e->hdr_gps_latitude_spin;
        DO_DESTROY_WIDGET
        w = &e->hdr_planetary_radius_spin;
        DO_DESTROY_WIDGET
        w = &e->hdr_cant_spin;
        DO_DESTROY_WIDGET
        w = &e->hdr_elevation_spin;
        DO_DESTROY_WIDGET
        w = &e->hdr_ground_is_water_tb;
	DO_DESTROY_WIDGET
        w = &e->hdr_ground_tile_width_spin;
	DO_DESTROY_WIDGET
        w = &e->hdr_ground_tile_height_spin;
	DO_DESTROY_WIDGET
        w = &e->hdr_ground_tile_close_range_spin;
	DO_DESTROY_WIDGET
        w = &e->hdr_ground_color_btn;
	DO_DESTROY_WIDGET
        w = &e->hdr_ground_texture_entry;
	DO_DESTROY_WIDGET
	w = &e->hdr_map_da;
	DO_DESTROY_WIDGET
	w = &e->hdr_map_width_spin;
	DO_DESTROY_WIDGET
	w = &e->hdr_map_height_spin;
	DO_DESTROY_WIDGET
        w = &e->hdr_sw_eb;
        DO_DESTROY_WIDGET
        w = &e->hdr_sw;
        DO_DESTROY_WIDGET

        w = &e->notebook;
        DO_DESTROY_WIDGET

        /* Delete tool bar. */
        ToolBarDelete(e->tool_bar);
        e->tool_bar = NULL;

	w = &e->menu_bar;
	DO_DESTROY_WIDGET
        w = &e->main_vbox;
        DO_DESTROY_WIDGET
	w = &e->toplevel;
	DO_DESTROY_WIDGET

	for(i = 0; i < e->total_opids; i++)
	{
	    editor_opid_struct *opid = e->opid[i];
	    if(opid == NULL)
		continue;
	    g_free(opid->tooltip);
	    g_free(opid);
	}
	g_free(e->opid);
	e->opid = NULL;
	e->total_opids = 0;

	pm = &e->view_pm;
	UNREF_PIXMAP

	pm = &e->list_object_pixmap;
	bm = &e->list_object_mask;
	UNREF_PIXMAP_MASK
        pm = &e->list_ground_pixmap;
        bm = &e->list_ground_mask;
	UNREF_PIXMAP_MASK
        pm = &e->list_helipad_pixmap;
        bm = &e->list_helipad_mask;
	UNREF_PIXMAP_MASK
        pm = &e->list_runway_pixmap;
        bm = &e->list_runway_mask;
	UNREF_PIXMAP_MASK
        pm = &e->list_human_pixmap;
        bm = &e->list_human_mask;
	UNREF_PIXMAP_MASK
        pm = &e->list_smoke_pixmap;
        bm = &e->list_smoke_mask;
        UNREF_PIXMAP_MASK
        pm = &e->list_fire_pixmap;
        bm = &e->list_fire_mask;
        UNREF_PIXMAP_MASK

        bm = &e->view_object_mask;
        UNREF_MASK
        bm = &e->view_ground_mask;
        UNREF_MASK
        bm = &e->view_helipad_mask;
        UNREF_MASK
        bm = &e->view_human_mask;
        UNREF_MASK
        bm = &e->view_smoke_mask;
        UNREF_MASK
        bm = &e->view_fire_mask;
        UNREF_MASK

	if(e->accelgrp != NULL)
	{
	    gtk_accel_group_unref(e->accelgrp);
	    e->accelgrp = NULL;
	}

	g_free(e->hdr_map_buf);
	e->hdr_map_buf = NULL;

	g_free(e->view_buf);
	e->view_buf = NULL;

	g_free(e);

#undef UNREF_PIXMAP_MASK
#undef DO_DESTROY_WIDGET
}
