#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/fio.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "cdialog.h"
#include "fb.h"
#include "csd.h"
#include "pdialog.h"
#include "toolbar.h"

#include "sfm.h"	/* For SFM math functions. */

#include "editor.h"
#include "editorcb.h"
#include "editorobj.h"
#include "editortex.h"
#include "editorlist.h"
#include "editorview.h"
#include "editorrun.h"
#include "editorscnopen.h"
#include "editorscnsave.h"
#include "editorcheck.h"
#include "scnedit.h"
#include "config.h"


void EditorNULLCB(GtkWidget *widget, gpointer data);
gint EditorDeleteEventCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
);
gint EditorScrolledWindowEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void EditorNotebookSwitchPageCB(
	GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
	gpointer data
);
void EditorCListSelectRowCB(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
);
void EditorCListUnselectRowCB(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
);
void EditorStopCB(GtkWidget *widget, gpointer data);

void EditorHdrEditableChangedCB(GtkWidget *widget, gpointer data);
void EditorHdrToggleButtonToggledCB(GtkWidget *widget, gpointer data);
void EditorHdrEditCListChangedCB(
        gpointer editclist, GtkCList *clist, gint row, gint column,
        gpointer data
);
void EditorHdrGroundColorCB(GtkWidget *widget, gpointer data);
void EditorHdrMapFileCB(GtkWidget *widget, gpointer data);
gint EditorHdrMapDrawingAreaEventCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
);

gint EditorViewExposeEventCB(
        GtkWidget *widget, GdkEventExpose *expose, gpointer data
);
gint EditorViewConfigureEventCB(
        GtkWidget *widget, GdkEventConfigure *configure, gpointer data
);
gint EditorViewFocusEventCB(
        GtkWidget *widget, GdkEventFocus *focus, gpointer data
);
gint EditorViewKeyEventCB(
        GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint EditorViewButtonEventCB(
        GtkWidget *widget, GdkEventButton *button, gpointer data
);
gint EditorViewMotionEventCB(
        GtkWidget *widget, GdkEventMotion *motion, gpointer data
);
static gint EditorViewScrollTOCB(gpointer data);
void EditorViewAdjValueChangedCB(
        GtkAdjustment *adj, gpointer data
);
gboolean EditorViewDragMotionCB(
        GtkWidget *widget, GdkDragContext *dc, gint x, gint y, guint t,
        gpointer data
);
void EditorViewDragDataReceivedCB(
        GtkWidget *widget, GdkDragContext *dc, gint x, gint y,
        GtkSelectionData *selection_data, guint info, guint t,
        gpointer data
);
void EditorViewUnitsChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
);
void EditorViewDisplayMapCB(
        GtkWidget *widget, gpointer data
);
void EditorViewDisplayObjectLabelsCB(
        GtkWidget *widget, gpointer data
);
void EditorViewDisplayPositionLabelCB(
        GtkWidget *widget, gpointer data
);
void EditorViewDisplayRegisteredLocationsCB(
        GtkWidget *widget, gpointer data
);

gint EditorObjListKeyEventCB(
        GtkWidget *widget, GdkEventKey *key, gpointer data
);
gint EditorObjListButtonEventCB(
        GtkWidget *widget, GdkEventButton *button, gpointer data
);
void EditorObjModelBrowseCB(gpointer widget, gpointer data);
void EditorObjTypeChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
);
void EditorObjPolygonOffsetToggledCB(gpointer widget, gpointer data);
void EditorObjSizeShapeChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
);
void EditorObjAdjValueChangedCB(
        GtkAdjustment *adj, gpointer data
);
void EditorObjReferenceAdjValueChangedCB(
        GtkAdjustment *adj, gpointer data
);
void EditorObjReferenceChangeCB(GtkWidget *widget, gpointer data);
void EditorObjPresetModelTypeChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
);

void EditorToolBarCB(gpointer data);
void EditorToolBarEnterCB(gpointer data);
void EditorToolBarLeaveCB(gpointer data);

void EditorNewCB(GtkWidget *widget, gpointer data);
void EditorOpenCB(GtkWidget *widget, gpointer data);
void EditorSaveCB(GtkWidget *widget, gpointer data);
void EditorSaveAsCB(GtkWidget *widget, gpointer data);
void EditorCloseCB(GtkWidget *widget, gpointer data);

void EditorCutCB(GtkWidget *widget, gpointer data);
void EditorCopyCB(GtkWidget *widget, gpointer data);
void EditorPasteCB(GtkWidget *widget, gpointer data);
void EditorNewObjectCB(GtkWidget *widget, gpointer data);
void EditorGoToSelectedCB(GtkWidget *widget, gpointer data);
void EditorMoveToCursorCB(GtkWidget *widget, gpointer data);
void EditorDeleteObjectCB(GtkWidget *widget, gpointer data);

void EditorViewHeaderCB(GtkWidget *widget, gpointer data);
void EditorViewSceneryCB(GtkWidget *widget, gpointer data);
void EditorRefreshCB(GtkWidget *widget, gpointer data);

void EditorCheckSceneryCB(GtkWidget *widget, gpointer data);
void EditorRunSARCB(GtkWidget *widget, gpointer data);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? (gfloat)atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define REMOVE_TOID(i)	{ if(*(i) != (guint)-1) { \
 gtk_timeout_remove(*(i)); *(i) = (guint)-1; } }


void EditorNULLCB(GtkWidget *widget, gpointer data)
{
	return;
}

/*
 *      Toplevel "delete_event" signal callback.
 */
gint EditorDeleteEventCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	EditorCloseCB(widget, data);
	return(TRUE);
}

/*
 *	Scrolled window GtkEventBox event signal callback.
 */
gint EditorScrolledWindowEventCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	gint etype;
	gboolean key_press = FALSE;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	GtkWidget *w;
	GtkScrolledWindow *sw;
	editor_struct *e = EDITOR(data);
	if((widget == NULL) || (event == NULL) || (e == NULL))
	    return(status);

	etype = *(gint *)event;

	if((etype != GDK_KEY_PRESS) &&
	   (etype != GDK_KEY_RELEASE)
	)
	{
	    if(widget->window != ((GdkEventAny *)event)->window)
		return(status);
	}

	/* Get GtkScrolledWindow based on the matching GtkEventBox. */
	if(widget == e->hdr_sw_eb)
	    w = e->hdr_sw;
	else if(widget == e->obj_sw_eb)
	    w = e->obj_sw;
	else
	    w = NULL;
	sw = (w != NULL) ? GTK_SCROLLED_WINDOW(w) : NULL;
	if(sw == NULL)
	    return(status);

	switch(etype)
	{
#define DO_STOP_KEY_SIGNAL				\
{ if(widget != NULL)					\
 gtk_signal_emit_stop_by_name(				\
  GTK_OBJECT(widget),					\
  key_press ? "key_press_event" : "key_release_event"	\
 );							\
}
	  case GDK_KEY_PRESS:
	    key_press = TRUE;
	  case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
	    e->drag_key_state = key->state;
	    switch(key->keyval)
	    {
              case GDK_Up:
              case GDK_KP_Up:
                if(key_press)
                {
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
                        sw->vscrollbar : NULL
                    );
                    GtkAdjustment *adj = (range != NULL) ?
                        range->adjustment : NULL;
                    if(adj != NULL)
                    {
                        adj->value -= (key->state & GDK_SHIFT_MASK) ?
                            adj->page_increment : adj->step_increment;
                        if(adj->value < adj->lower)
                            adj->value = adj->lower;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
                }
		DO_STOP_KEY_SIGNAL
                status = TRUE;
                break;

              case GDK_Down:
              case GDK_KP_Down:
                if(key_press)
                {
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
                        sw->vscrollbar : NULL
                    );
                    GtkAdjustment *adj = (range != NULL) ?
                        range->adjustment : NULL;
                    if(adj != NULL)
                    {
                        adj->value += (key->state & GDK_SHIFT_MASK) ?
			    adj->page_increment : adj->step_increment;
                        if(adj->value > (adj->upper - adj->page_size))
                            adj->value = adj->upper - adj->page_size;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
                }
		DO_STOP_KEY_SIGNAL
                status = TRUE;
                break;

              case GDK_Left:
              case GDK_KP_Left:
                if(key_press)
                {
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
                        sw->hscrollbar : NULL
                    );
                    GtkAdjustment *adj = (range != NULL) ?
                        range->adjustment : NULL;
                    if(adj != NULL)
                    {
                        adj->value -= (key->state & GDK_SHIFT_MASK) ?
                            adj->page_increment : adj->step_increment;
                        if(adj->value < adj->lower)
                            adj->value = adj->lower;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
                }
		DO_STOP_KEY_SIGNAL
                status = TRUE;
                break;

              case GDK_Right:
              case GDK_KP_Right:
                if(key_press)
                {
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
                        sw->hscrollbar : NULL
                    );
                    GtkAdjustment *adj = (range != NULL) ?
                        range->adjustment : NULL;
                    if(adj != NULL)
                    {
                        adj->value += (key->state & GDK_SHIFT_MASK) ?
                            adj->page_increment : adj->step_increment;
                        if(adj->value > (adj->upper - adj->page_size))
                            adj->value = adj->upper - adj->page_size;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
                }
		DO_STOP_KEY_SIGNAL
                status = TRUE;
                break;

              case '-': case 'b':
              case GDK_Page_Up:
              case GDK_KP_Page_Up:
                if(key_press)
                {
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
                        sw->vscrollbar : NULL
                    );
                    GtkAdjustment *adj = (range != NULL) ?
                        range->adjustment : NULL;
                    if(adj != NULL)
                    {
                        adj->value -= adj->page_increment;
                        if(adj->value < adj->lower)
                            adj->value = adj->lower;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
                }
                status = TRUE;
                break;

	      case GDK_space:
	      case GDK_Page_Down:
              case GDK_KP_Page_Down:
		if(key_press)
		{
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
			sw->vscrollbar : NULL
		    );
                    GtkAdjustment *adj = (range != NULL) ?
			range->adjustment : NULL;
                    if(adj != NULL)
                    {
			adj->value += adj->page_increment;
                        if(adj->value > (adj->upper - adj->page_size))
                            adj->value = adj->upper - adj->page_size;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
		}
		status = TRUE;
		break;

              case GDK_Home:
              case GDK_KP_Home:
                if(key_press)
                {
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
                        sw->vscrollbar : NULL
                    );
                    GtkAdjustment *adj = (range != NULL) ?
                        range->adjustment : NULL;
                    if(adj != NULL)
                    {
                        adj->value = adj->lower;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
                }
                status = TRUE;
                break;

              case GDK_End:
              case GDK_KP_End:
                if(key_press)
                {
                    GtkRange *range = (GtkRange *)((sw != NULL) ?
                        sw->vscrollbar : NULL
                    );
                    GtkAdjustment *adj = (range != NULL) ?
                        range->adjustment : NULL;
                    if(adj != NULL)
                    {
                        adj->value = adj->upper - adj->page_size;
                        gtk_signal_emit_by_name(
                            GTK_OBJECT(adj), "value_changed"
                        );
                    }
                }
                status = TRUE;
                break;
	    }
	    break;
#undef DO_STOP_KEY_SIGNAL

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    if(!GTK_WIDGET_HAS_FOCUS(widget))
		gtk_widget_grab_focus(widget);
	    e->drag_button = button->button;
	    e->drag_last_x = (gint)button->x;
            e->drag_last_y = (gint)button->y;
	    switch(button->button)
	    {
	      case 1: case 2:
		if(TRUE)
		{
		    se_core_struct *core_ptr = SE_CORE(e->core_ptr);
                    gdk_pointer_grab(
                        widget->window,
                        FALSE,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                            GDK_POINTER_MOTION_MASK,
                        NULL,
                        (core_ptr != NULL) ?
                            core_ptr->cursor_translate : NULL,
                        button->time
                    );
		}
		status = TRUE;
		break;

	      case 3:
		if((widget == e->obj_sw_eb) && (e->obj_menu != NULL))
		{
		    GtkMenu *menu = GTK_MENU(e->obj_menu);
		    gtk_menu_popup(
                        menu, NULL, NULL,
                        NULL, NULL,
                        button->button, button->time
		    );
                }
                status = TRUE;
                break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;
	    e->drag_button = 0;
	    if(gdk_pointer_is_grabbed())
	    {
		gdk_pointer_ungrab(button->time);
		status = TRUE;
	    }
	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if((e->drag_button == 1) || (e->drag_button == 2))
	    {
		gint x, y, dx, dy;
		GdkModifierType mask;
		GtkRange *range;
		GtkAdjustment *hadj, *vadj;

		range = (GtkRange *)((sw != NULL) ? sw->hscrollbar : NULL);
		hadj = (range != NULL) ? range->adjustment : NULL;
		range = (GtkRange *)((sw != NULL) ? sw->vscrollbar : NULL);
                vadj = (range != NULL) ? range->adjustment : NULL;

		/* Get pointer position. */
		if(motion->is_hint)
		{
		    gdk_window_get_pointer(
			motion->window, &x, &y, &mask
		    );
		}
		else
		{
		    x = (gint)motion->x;
		    y = (gint)motion->y;
		}

		dx = x - e->drag_last_x;
                dy = y - e->drag_last_y;

		if((hadj != NULL) && (dx != 0))
		{
		    hadj->value -= (gfloat)dx;
		    if(hadj->value > (hadj->upper - hadj->page_size))
			hadj->value = hadj->upper - hadj->page_size;
		    if(hadj->value < hadj->lower)
			hadj->value = hadj->lower;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(hadj), "value_changed"
		    );
		}
                if((vadj != NULL) && (dy != 0))
                {
                    vadj->value -= (gfloat)dy;
                    if(vadj->value > (vadj->upper - vadj->page_size))
                        vadj->value = vadj->upper - vadj->page_size;
                    if(vadj->value < vadj->lower)
                        vadj->value = vadj->lower;
                    gtk_signal_emit_by_name(
                        GTK_OBJECT(vadj), "value_changed"
                    );
                }

		status = TRUE;
	    }
            break;



	}

	return(status);
}

/*
 *	GtkNotebook "switch_page" callback.
 */
void EditorNotebookSwitchPageCB(
        GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
        gpointer data
)
{
	editor_struct *e = EDITOR(data);
	if((notebook == NULL) || (e == NULL))
	    return;

	if(notebook == (GtkNotebook *)EDITOR_NOTEBOOK(e))
	{
	    e->page_num = page_num;
	    EditorUpdateMenus(e);
	}
}

/*
 *	CList "select_row" signal callback.
 */
void EditorCListSelectRowCB(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
)
{
	editor_struct *e = EDITOR(data);
        if((clist == NULL) || (e == NULL))
            return;

        if(gtk_clist_row_is_visible(clist, row) !=
            GTK_VISIBILITY_FULL
        )
            gtk_clist_moveto(
                clist,
                row, -1,        /* Row, column. */
                0.5, 0.0        /* Row, column. */
            );

	if(GTK_WIDGET(clist) == e->obj_clist)
	{
	    gchar *buf;
	    GList *glist = clist->selection;
	    const editor_obj_struct *o = EDITOR_OBJ(
		gtk_clist_get_row_data(clist, row)
	    );

	    /* If exactly one object selected then get values. */
	    if((glist != NULL) ? (g_list_length(glist) == 1) : FALSE)
	    {
		e->obj_selected = row;
		EditorObjListDoGetValues(e, row, (const gpointer)o);
	    }
	    else
	    {
		e->obj_selected = -1;
		EditorObjListDoGetValues(e, -1, NULL);
	    }

	    /* Format status bar message. */
	    if((o != NULL) ? (o->name == NULL) : TRUE)
		buf = g_strdup_printf("Selected object #%i", row);
	    else
		buf = g_strdup_printf("Selected object \"%s\"", o->name);

	    /* Redraw view and set status bar message. */
	    EditorViewDraw(e);
	    StatusBarMessage(e->status_bar, buf, FALSE);
	    g_free(buf);
	}

	EditorUpdateMenus(e);
}

/*
 *	CList "unselect_row" signal callback.
 */
void EditorCListUnselectRowCB(
        GtkCList *clist, gint row, gint column, GdkEvent *event,
        gpointer data
)
{
        editor_struct *e = EDITOR(data);
        if((clist == NULL) || (e == NULL))
            return;

        if(GTK_WIDGET(clist) == e->obj_clist)
	{
	    gint on = e->obj_selected;
	    if(on > -1)
	    {
		editor_obj_struct *o = EDITOR_OBJ(
		    gtk_clist_get_row_data(clist, on)
		);
		if(o != NULL)
		{
		    EditorObjListDoApplyValues(e, o);
		    EditorObjListRealizeAll(e, TRUE);
		}
		e->obj_selected = on = -1;
	    }
	    EditorObjListDoGetValues(e, -1, NULL);
            EditorViewDraw(e);
	}

        EditorUpdateMenus(e);
}

/*
 *	Editor stop callback.
 */
void EditorStopCB(GtkWidget *widget, gpointer data)
{
	editor_struct *e = EDITOR(data);
	if(e == NULL)
	    return;

	e->stop_count++;
}


/*
 *	Header GtkEditable "changed" signal callback.
 */
void EditorHdrEditableChangedCB(GtkWidget *widget, gpointer data)
{
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	if(e->hdr_freeze_count <= 0)
	{
	    if(!EDITOR_HAS_CHANGES(e))
	    {
		e->has_changes = TRUE;
		EditorUpdateTitle(e);
		EditorUpdateMenus(e);
	    }
	}
}

/*
 *	Header GtkToggleButton "toggled" signal callback.
 */
void EditorHdrToggleButtonToggledCB(GtkWidget *widget, gpointer data)
{
	EditorHdrEditableChangedCB(NULL, data);
}

/*
 *	Header Editable CList changed signal callback.
 */
void EditorHdrEditCListChangedCB(
        gpointer editclist, GtkCList *clist, gint row, gint column,
        gpointer data
)
{
	EditorHdrEditableChangedCB(NULL, data);
}

/*
 *      Ground color select callback.
 */
void EditorHdrGroundColorCB(GtkWidget *widget, gpointer data)
{
        gboolean status;
        GtkWidget *w;
        csd_color_struct c, *c_rtn = NULL;
	void (*color_changed_cb)(gpointer, csd_color_struct *) =
	    (gpointer)CSDColorButtonSetColor;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(CSDIsQuery())
            return;

        EditorSetBusy(e, TRUE);

        w = e->hdr_ground_color_btn;
	CSDColorButtonGetColor(w, &c);

        /* Get response. */
        CSDSetTransientFor(EDITOR_TOPLEVEL(e));
        status = CSDGetResponse(
            "Select Color",
            "Select", "Cancel",
            &c, &c_rtn,
            w, color_changed_cb
        );
        CSDSetTransientFor(NULL);

        /* Got response? */
        if(status && (c_rtn != NULL))
        {
	    CSDColorButtonSetColor(w, c_rtn);
	    if(!EDITOR_HAS_CHANGES(e))
	    {
		e->has_changes = TRUE;
		EditorUpdateTitle(e);
	    }
	    EditorUpdateMenus(e);
        }
        else
        {
	    CSDColorButtonSetColor(w, &c);
        }

        EditorSetBusy(e, FALSE);
}

/*
 *	Header map file callback.
 */
void EditorHdrMapFileCB(GtkWidget *widget, gpointer data)
{
        gboolean status;
        const gchar *s, *cwd;
        gchar *filename, *filename_short;
        fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
        gint total_ftypes = 0;
        gchar **path_rtn = NULL;
        gint total_path_rtns = 0;
	const editor_tex_struct *t;
	se_core_struct *core_ptr;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(FileBrowserIsQuery())
            return;

	core_ptr = SE_CORE(e->core_ptr);

        EditorSetBusy(e, TRUE);

	/* Get current working directory. */
	cwd = (core_ptr != NULL) ? core_ptr->cwd : NULL;
	if(cwd == NULL)
#ifdef __MSW__
	    cwd = "\\";
#else
	    cwd = "/";
#endif

        /* Create file types list. */
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            ".tex .tga", "Vertex 3D Texture files"
        );
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            "*.*", "All files"
        );

	/* Get initial path. */
	t = e->hdr_map_tex;
	if((t != NULL) ? (t->filename != NULL) : FALSE)
	{
	    if(ISPATHABSOLUTE(t->filename))
		filename = STRDUP(t->filename);
	    else
		filename = g_strdup_printf(
		    "%s%c%s",
		    SE_DEF_SAR_DATA_DIR, DIR_DELIMINATOR,
		    t->filename
		);
	}
	else
	{
#ifdef __MSW__
	    filename = STRDUP(cwd);
#else
	    filename = STRDUP(SE_DEF_SAR_DATA_DIR);
#endif
	}

        /* Query user for map file. */
        FileBrowserSetTransientFor(EDITOR_TOPLEVEL(e));
        status = FileBrowserGetResponse(
            "Select Map File",
            "Select", "Cancel",
            filename,
            ftype, total_ftypes,
            &path_rtn, &total_path_rtns,
            &ftype_rtn
        );
        FileBrowserSetTransientFor(NULL);
        FileBrowserDeleteTypeList(ftype, total_ftypes);
	g_free(filename);

        /* Get filename if there was a response. */
        s = (status && (total_path_rtns > 0)) ?
            path_rtn[0] : NULL;
        filename = STRDUP(s);
        if(filename == NULL)
        {
            EditorSetBusy(e, FALSE);
            return;
        }

	/* Get shortened filename if the filename has a prefix of the
	 * data directory.
	 */
#ifdef __MSW__
        if(strpfx(filename, cwd))
        {
            s = filename + strlen(cwd);
            while(*s == DIR_DELIMINATOR)
                s++;
        }
#else
        if(strpfx(filename, SE_DEF_SAR_DATA_DIR))
        {
            s = filename + strlen(SE_DEF_SAR_DATA_DIR);
            while(*s == DIR_DELIMINATOR)
                s++;
        }
#endif
        else
        {
            s = filename;
        }
        filename_short = STRDUP(s);

        /* Set new map file. */
	EditorTexDelete(e->hdr_map_tex);
	e->hdr_map_tex = EditorTexNewLoad(
	    filename_short,
	    NULL,
	    filename_short,
	    1.0f
	);

	g_free(filename_short);
        g_free(filename);

        EditorSetBusy(e, FALSE);

	if(!EDITOR_HAS_CHANGES(e))
	{
	    e->has_changes = TRUE;
	    EditorUpdateTitle(e);
	}

	EditorUpdateMenus(e);
	EditorHdrMapDraw(e);
        EditorViewDraw(e);
}

/*
 *	Header map GtkDrawingArea event callback. 
 */
gint EditorHdrMapDrawingAreaEventCB(
        GtkWidget *widget, GdkEvent *event, gpointer data
)
{
        gint status = FALSE;
        gint etype;
        GdkEventConfigure *configure;
        GdkEventExpose *expose;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	GdkEventCrossing *crossing;
	GdkEventFocus *focus;
	GdkWindow *window;
        editor_struct *e = EDITOR(data);
        if((widget == NULL) || (event == NULL) || (e == NULL))
            return(status);

	window = widget->window;
        if(window != ((GdkEventAny *)event)->window)
            return(status);

        etype = *(gint *)event;
        switch(etype)
        {
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    if((configure->width > 0) && (configure->height > 0))
	    {
		g_free(e->hdr_map_buf);
		e->hdr_map_buf = (guint8 *)g_malloc(
		    configure->width * configure->height * 4
		);
		EditorHdrMapDraw(e);
	    }
	    status = TRUE;
	    break;

	  case GDK_EXPOSE:
	    expose = (GdkEventExpose *)event;
	    EditorHdrMapDraw(e);
	    status = TRUE;
	    break;

	  case GDK_KEY_PRESS:
	    key = (GdkEventKey *)event;
	    e->drag_key_state = key->state;
	    status = TRUE;
            break;

          case GDK_KEY_RELEASE:
            key = (GdkEventKey *)event;
            status = TRUE;
            break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    e->drag_button = button->button;
            e->drag_key_state = button->state;
            e->drag_last_x = (gint)button->x;
            e->drag_last_y = (gint)button->y;
	    if(!GTK_WIDGET_HAS_FOCUS(widget))
		gtk_widget_grab_focus(widget);
	    status = TRUE;
            break;

          case GDK_BUTTON_RELEASE:
            button = (GdkEventButton *)event;
	    e->drag_button = 0;
	    if(gdk_pointer_is_grabbed())
		gdk_pointer_ungrab(button->time);
	    switch(button->button)
	    {
	      case 1:
		EditorHdrMapFileCB(NULL, data);
		break;
	    }
            status = TRUE;
            break;

          case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if(e->drag_button == 0)
	    {
                if(e->hdr_map_tex != NULL)
                {
                    const editor_tex_struct *t = e->hdr_map_tex;
                    if(t->filename != NULL)
                    {
			gchar *buf = g_strdup_printf(
			    "%s (%ix%i)",
			    t->filename, t->width, t->height
			);
                        GUISetWidgetTip(widget, buf);
                        GUIShowTipsNow(widget);
			g_free(buf);
		    }
                }
            }
            status = TRUE;
            break;

	  case GDK_ENTER_NOTIFY:
	    crossing = (GdkEventCrossing *)event;
            status = TRUE;
            break;

          case GDK_LEAVE_NOTIFY:
            crossing = (GdkEventCrossing *)event;
            status = TRUE;
            break;

	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    if(focus->in)
		GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
	    else
		GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
	    EditorHdrMapDraw(e);
            status = TRUE;
            break;
	}

	return(status);
}

/*
 *      View "expose_event" signal callback.
 */
gint EditorViewExposeEventCB(
        GtkWidget *widget, GdkEventExpose *expose, gpointer data
)
{
        gint status = FALSE;
        editor_struct *e = EDITOR(data);
        if((expose == NULL) || (e == NULL))
            return(status);

	EditorViewDraw(e);

	return(status);
}

/*
 *	View "configure_event" signal callback.
 */
gint EditorViewConfigureEventCB(
        GtkWidget *widget, GdkEventConfigure *configure, gpointer data
)
{
	gint status = FALSE;
	editor_struct *e = EDITOR(data);
	if((configure == NULL) || (e == NULL))
	    return(status);

	if((configure->width > 0) && (configure->height > 0))
	{
	    if(e->view_pm != NULL)
		gdk_pixmap_unref(e->view_pm);
	    e->view_pm = gdk_pixmap_new(
		(widget != NULL) ? widget->window : NULL,
		configure->width, configure->height, -1
	    );

	    g_free(e->view_buf);
	    e->view_buf = (guint8 *)g_malloc(
		configure->width * configure->height * 4
	    );
	}

        return(status);
}

/*
 *	View "focus_in_event" signal callback.
 */
gint EditorViewFocusEventCB(
        GtkWidget *widget, GdkEventFocus *focus, gpointer data
)
{
        gint status = FALSE;
        editor_struct *e = EDITOR(data);
        if((widget == NULL) || (focus == NULL) || (e == NULL))
            return(status);

	if(focus->in)
            GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
        else
            GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);

	if(widget == e->view_da)
	    EditorViewDraw(e);

	return(status);
}

/*
 *	View "key_press_event" signal callback.
 */
gint EditorViewKeyEventCB(
        GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
        gint status = FALSE;
        gboolean key_press;
	GtkAdjustment *adj;
	GtkWidget *w;
        editor_struct *e = EDITOR(data);
        if((key == NULL) || (e == NULL))
            return(status);

        key_press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;

#define DO_STOP_KEY_SIGNAL      \
{ if(widget != NULL) \
 gtk_signal_emit_stop_by_name( \
  GTK_OBJECT(widget), \
  key_press ? "key_press_event" : "key_release_event" \
 ); \
}

	e->drag_key_state = key->state;

        switch(key->keyval)
        {
          case GDK_Return:
          case GDK_KP_Enter:
          case GDK_ISO_Enter:
          case GDK_3270_Enter:

            break;

	  case GDK_Up:
          case GDK_KP_Up:
	    if(key_press)
	    {
		gint height;
		gfloat zoom;
		EditorViewGetViewZoomM(e, &zoom);
		w = e->view_da;
		height = (w != NULL) ? w->allocation.height : 0;
		w = e->view_y_spin;
		adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
		if((adj != NULL) && (height > 0) && (zoom > 0.0f))
		{
		    gfloat ym = ((key->state & GDK_CONTROL_MASK) ?
			0.5f * height : 0.1f * height) * zoom / height;
		    /* Convert to displayed units. */
                    switch(EDITOR_UNITS(e))
                    {
                      case EDITOR_UNITS_METRIC:
                      case EDITOR_UNITS_METRIC_ALT_FEET:
                        break;
                      case EDITOR_UNITS_ENGLISH:
                        ym = (gfloat)SFMMetersToFeet(ym);
                        break;
                    }
		    adj->value += ym;
		    gtk_signal_emit_by_name(
			GTK_OBJECT(adj), "value_changed"
		    );
		}
	    }
	    DO_STOP_KEY_SIGNAL
	    status = TRUE;
	    break;

          case GDK_Down:
          case GDK_KP_Down:
            if(key_press)
            {
                gint height;
                gfloat zoom;
                EditorViewGetViewZoomM(e, &zoom);
                w = e->view_da;
                height = (w != NULL) ? w->allocation.height : 0;
                w = e->view_y_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                if((adj != NULL) && (height > 0) && (zoom > 0.0f))
                {
                    gfloat ym = ((key->state & GDK_CONTROL_MASK) ?
                        0.5f * height : 0.1f * height) * zoom / height;
                    /* Convert to displayed units. */
                    switch(EDITOR_UNITS(e))
                    {
                      case EDITOR_UNITS_METRIC:
                      case EDITOR_UNITS_METRIC_ALT_FEET:
                        break;
                      case EDITOR_UNITS_ENGLISH:
                        ym = (gfloat)SFMMetersToFeet(ym);
                        break;
                    }
                    adj->value -= ym;
                    gtk_signal_emit_by_name(
                        GTK_OBJECT(adj), "value_changed"
                    );
                }
            }
            DO_STOP_KEY_SIGNAL
            status = TRUE;
            break;

          case GDK_Left:
          case GDK_KP_Left:
            if(key_press)
            {
                gint width;
                gfloat zoom;
                EditorViewGetViewZoomM(e, &zoom);
                w = e->view_da;
                width = (w != NULL) ? w->allocation.width : 0;
                w = e->view_x_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                if((adj != NULL) && (width > 0) && (zoom > 0.0f))
                {
                    gfloat xm = ((key->state & GDK_CONTROL_MASK) ?
                        0.5f * width : 0.1f * width) * zoom / width;
                    /* Convert to displayed units. */
                    switch(EDITOR_UNITS(e))
                    {
                      case EDITOR_UNITS_METRIC:
                      case EDITOR_UNITS_METRIC_ALT_FEET:
                        break;
                      case EDITOR_UNITS_ENGLISH:
                        xm = (gfloat)SFMMetersToFeet(xm);
                        break;
                    }
                    adj->value -= xm;
                    gtk_signal_emit_by_name(
                        GTK_OBJECT(adj), "value_changed"
                    );
                }
            }
            DO_STOP_KEY_SIGNAL
            status = TRUE;
            break;

          case GDK_Right:
          case GDK_KP_Right:
            if(key_press)
            {
                gint width;
                gfloat zoom;
                EditorViewGetViewZoomM(e, &zoom);
                w = e->view_da;
                width = (w != NULL) ? w->allocation.width : 0;
                w = e->view_x_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                if((adj != NULL) && (width > 0) && (zoom > 0.0f))
                {
                    gfloat xm = ((key->state & GDK_CONTROL_MASK) ?
                        0.5f * width : 0.1f * width) * zoom / width;
                    /* Convert to displayed units. */
                    switch(EDITOR_UNITS(e))
                    {
                      case EDITOR_UNITS_METRIC:
                      case EDITOR_UNITS_METRIC_ALT_FEET:
                        break;
                      case EDITOR_UNITS_ENGLISH:
                        xm = (gfloat)SFMMetersToFeet(xm);
                        break;
                    }
                    adj->value += xm;
                    gtk_signal_emit_by_name(
                        GTK_OBJECT(adj), "value_changed"
                    );
                }
            }
            DO_STOP_KEY_SIGNAL
            status = TRUE;
            break;

          case GDK_Page_Up:
          case GDK_KP_Page_Up:
	  case '-': case '_':
            if(key_press)
            {
                gint height;
                gfloat zoom;
                EditorViewGetViewZoomM(e, &zoom);
                w = e->view_da;
                height = (w != NULL) ? w->allocation.height : 0;
                w = e->view_zoom_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                if((adj != NULL) && (height > 0) && (zoom > 0.0f))
		{
                    gfloat ym = ((key->state & GDK_CONTROL_MASK) ?
                        0.5f * height : 0.1f * height) * zoom / height;
                    /* Convert to displayed units. */
                    switch(EDITOR_UNITS(e))
                    {
                      case EDITOR_UNITS_METRIC:
                      case EDITOR_UNITS_METRIC_ALT_FEET:
                        break;
                      case EDITOR_UNITS_ENGLISH:
                        ym = (gfloat)SFMMetersToFeet(ym);
                        break;
                    }
                    adj->value += ym;
		    if(adj->value > (adj->upper - adj->page_size))
			adj->value = adj->upper - adj->page_size;
                    gtk_signal_emit_by_name(
                        GTK_OBJECT(adj), "value_changed"
                    );
		}
	    }
            status = TRUE;
            break;

          case GDK_Page_Down:
          case GDK_KP_Page_Down:
	  case '=': case '+':
            if(key_press)
            {
                gint height;
                gfloat zoom;
                EditorViewGetViewZoomM(e, &zoom);
                w = e->view_da;
                height = (w != NULL) ? w->allocation.height : 0;
                w = e->view_zoom_spin;
                adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
                if((adj != NULL) && (height > 0) && (zoom > 0.0f))
                {
                    gfloat ym = ((key->state & GDK_CONTROL_MASK) ?
                        0.5f * height : 0.1f * height) * zoom / height;
                    /* Convert to displayed units. */
                    switch(EDITOR_UNITS(e))
                    {
                      case EDITOR_UNITS_METRIC:
                      case EDITOR_UNITS_METRIC_ALT_FEET:
                        break;
                      case EDITOR_UNITS_ENGLISH:
                        ym = (gfloat)SFMMetersToFeet(ym);
                        break;
                    }
                    adj->value -= ym;
                    if(adj->value < adj->lower)
                        adj->value = adj->lower;
                    gtk_signal_emit_by_name(
                        GTK_OBJECT(adj), "value_changed"
                    );
                }
            }
            status = TRUE;
            break;

          case GDK_Insert:
            if(key_press)
            {
                if(key->state & GDK_CONTROL_MASK)
                    EditorCopyCB(NULL, e);
                else if(key->state & GDK_SHIFT_MASK)
                    EditorPasteCB(NULL, e);
                else if(key->state == 0x00000000)
                    EditorNewObjectCB(NULL, e);
            }
            status = TRUE;
            break;

          case GDK_Delete:
            if(key_press)
            {
                EditorDeleteObjectCB(NULL, e);
            }
            status = TRUE;
            break;
	}

#undef DO_STOP_KEY_SIGNAL

	return(status);
}

/*
 *	View "button_press_event" signal callback.
 */
gint EditorViewButtonEventCB(
        GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
        gint status = FALSE;
	se_core_struct *core_ptr;
	editor_struct *e = EDITOR(data);
        if((button == NULL) || (e == NULL))
            return(status);

	core_ptr = SE_CORE(e->core_ptr);

        switch((gint)button->type)
        {
          case GDK_BUTTON_PRESS:
	    e->drag_button = button->button;
	    e->drag_key_state = button->state;
	    e->drag_last_x = (gint)button->x;
            e->drag_last_y = (gint)button->y;
	    if(!GTK_WIDGET_HAS_FOCUS(widget))
		gtk_widget_grab_focus(widget);
	    switch(button->button)
	    {
	      case 1:
                /* Rectangular select? */
                if(e->drag_key_state & GDK_SHIFT_MASK)
                {
                    e->view_rect_select = TRUE;
                    e->view_rect_select_x0 =
                        e->view_rect_select_x1 = (gint)button->x;
                    e->view_rect_select_y0 =
                        e->view_rect_select_y1 = (gint)button->y;
                    gdk_pointer_grab(
                        (widget != NULL) ? widget->window : NULL,
                        FALSE,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                            GDK_POINTER_MOTION_MASK,
                        NULL,
                        (core_ptr != NULL) ?
                            core_ptr->cursor_rect_select : NULL,
                        button->time
                    );
                }
		/* Translate? */
		else if(e->drag_key_state & GDK_CONTROL_MASK)
		{
                    gdk_pointer_grab(
                        (widget != NULL) ? widget->window : NULL,
                        FALSE,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                            GDK_POINTER_MOTION_MASK,
                        NULL,
                        (core_ptr != NULL) ?
			    core_ptr->cursor_translate : NULL,
                        button->time
                    );
		}
		/* Zoom? */
		else if(e->drag_key_state & GDK_MOD1_MASK)
		{
                    gdk_pointer_grab(
                        (widget != NULL) ? widget->window : NULL,
                        FALSE,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                            GDK_POINTER_MOTION_MASK,
                        NULL,
                        (core_ptr != NULL) ?
                            core_ptr->cursor_zoom : NULL,
                        button->time
                    );
		}
		/* Set cursor position? */
                else
		{
		    gfloat xm, ym, zm;

		    /* Update cursor position. */
		    EditorViewGetCursorM(e, NULL, NULL, &zm);
		    EditorViewGetPositionWToM(
			e, (gint)button->x, (gint)button->y, &xm, &ym
		    );
		    EditorViewSetCursorM(e, xm, ym, zm);
		    EditorViewAdjValueChangedCB(NULL, e);

                    gdk_pointer_grab(
                        (widget != NULL) ? widget->window : NULL,
                        FALSE,
                        GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                            GDK_POINTER_MOTION_MASK,
                        NULL,
			NULL,
                        button->time
                    );
		}
		break;

              case 2:
		/* Zoom. */
                gdk_pointer_grab(
                    (widget != NULL) ? widget->window : NULL,
                    FALSE,
                    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                        GDK_POINTER_MOTION_MASK,
                    NULL,
                    (core_ptr != NULL) ?
                        core_ptr->cursor_zoom : NULL,
                    button->time
                );
                break;

	      case 3:
                if(e->obj_menu != NULL)
                {
                    GtkMenu *menu = GTK_MENU(e->obj_menu);
                    gtk_menu_popup(
                        menu, NULL, NULL,
                        NULL, NULL,
                        button->button, button->time
                    );
                }
		break;
	    }
	    status = TRUE;
	    break;

          case GDK_BUTTON_RELEASE:
            e->drag_button = 0;
	    if(gdk_pointer_is_grabbed())
		gdk_pointer_ungrab(button->time);

	    /* Remove view scroll timeout (if any). */
	    REMOVE_TOID(&e->view_scroll_toid);

	    /* Was performing a rectangular select? */
	    if(e->view_rect_select)
	    {
		EditorViewProcessRectangularSelect(e);
		e->view_rect_select = FALSE;
                EditorViewAdjValueChangedCB(NULL, e);
	    }

            status = TRUE;
            break;

	}

	return(status);
}

/*
 *      View "motion_notify_event" signal callback.
 */
gint EditorViewMotionEventCB(
        GtkWidget *widget, GdkEventMotion *motion, gpointer data
)
{
        gint status = FALSE;
	gint x, y, width, height, dx, dy;
	gfloat zoom, dxm, dym;
	GdkModifierType mask;
	GdkWindow *window;
	GtkAdjustment *adj;
	GtkWidget *w;
        editor_struct *e = EDITOR(data);
        if((motion == NULL) || (e == NULL))
            return(status);

        /* Get view widget and make sure it is mapped and realized. */
        w = e->view_da;
        if((w != NULL) ?
            (!GTK_WIDGET_MAPPED(w) || !GTK_WIDGET_REALIZED(w))
            : TRUE
        )
            return(status);

	/* Get view window. */
	window = w->window;
	if(window == NULL)
	    return(status);

	/* Get size of window. */
	gdk_window_get_size(window, &width, &height);
	if((width <= 0) || (height <= 0))
	    return(status);

	/* Get pointer position. */
	if(motion->is_hint && (e->drag_button > 0))
	{
	    gdk_window_get_pointer(
		motion->window, &x, &y, &mask
	    );
	}
	else
	{
	    x = (gint)motion->x;
	    y = (gint)motion->y;
	}

	/* Get pointer delta change. */
	dx = x - e->drag_last_x;
        dy = y - e->drag_last_y;

	/* Update pointer position. */
        e->drag_last_x = x;
	e->drag_last_y = y;


        /* Rectangular select */
        if((e->drag_button == 1) &&
           (e->drag_key_state & GDK_SHIFT_MASK)
        )
        {
            e->view_rect_select_x1 = x;
            e->view_rect_select_y1 = y;
            EditorViewAdjValueChangedCB(NULL, e);
            status = TRUE;
        }
	/* Translating? */
	else if((e->drag_button == 1) &&
                (e->drag_key_state & GDK_CONTROL_MASK)
	)
	{
	    /* Calculate change in meters. */
	    EditorViewGetViewZoomM(e, &zoom);
	    dxm = (gfloat)dx * zoom / (gfloat)width;
	    dym = (gfloat)-dy * zoom / (gfloat)height;

	    /* Convert dxm and dym to correct units. */
	    switch(EDITOR_UNITS(e))
	    {
	      case EDITOR_UNITS_METRIC:
              case EDITOR_UNITS_METRIC_ALT_FEET:
		break;
	      case EDITOR_UNITS_ENGLISH:
		dxm = (gfloat)SFMMetersToFeet(dxm);
                dym = (gfloat)SFMMetersToFeet(dym);
		break;
	    }

	    /* Update position. */
	    if((dxm != 0.0f) || (dym != 0.0f))
	    {
		e->view_freeze_count++;

		w = e->view_x_spin;
		adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
		if((adj != NULL) && (dxm != 0.0f))
		{
		    adj->value = CLIP(
			adj->value - dxm,
			adj->lower, adj->upper - adj->page_size
		    );
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}
		w = e->view_y_spin;
		adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
		if((adj != NULL) && (dym != 0.0f))
		{
		    adj->value = CLIP(
			adj->value - dym,
			adj->lower, adj->upper - adj->page_size
		    );
		    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
		}

		e->view_freeze_count--;
		EditorViewAdjValueChangedCB(NULL, e);
	    }
	    status = TRUE;
	}
        /* Zooming? */
	else if(((e->drag_button == 1) &&
                 (e->drag_key_state & GDK_MOD1_MASK)) ||
                (e->drag_button == 2)
        )
	{
	    /* Calculate change in meters. */
            EditorViewGetViewZoomM(e, &zoom);
            dym = (gfloat)-dy * zoom / (gfloat)height;

            /* Convert dxm and dym to correct units. */
            switch(EDITOR_UNITS(e))
            {
              case EDITOR_UNITS_METRIC:
              case EDITOR_UNITS_METRIC_ALT_FEET:
                break;
              case EDITOR_UNITS_ENGLISH:
                dym = (gfloat)SFMMetersToFeet(dym);
                break;
            }

            e->view_freeze_count++;

	    /* Update zoom. */
            w = e->view_zoom_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if(adj != NULL)
            {
                adj->value += dym;
		if(adj->value > (adj->upper - adj->page_size))
		    adj->value = adj->upper - adj->page_size;
                if(adj->value < adj->lower)
                    adj->value = adj->lower;
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }

            e->view_freeze_count--;
	    EditorViewAdjValueChangedCB(NULL, e);
	    status = TRUE;
	}
        /* Dragging cursor? */
        else if(e->drag_button == 1)
	{
            gfloat xm, ym, zm;

	    /* Update cursor position. */
            EditorViewGetCursorM(e, NULL, NULL, &zm);
            EditorViewGetPositionWToM(
                e, (gint)motion->x, (gint)motion->y, &xm, &ym
            );
	    EditorViewSetCursorM(e, xm, ym, zm);
	    EditorViewAdjValueChangedCB(NULL, e);

	    /* Check if pointer moved off the window. */
	    if(((gint)motion->x < 0) || ((gint)motion->x >= width) ||
               ((gint)motion->y < 0) || ((gint)motion->y >= height)
	    )
	    {
		if(e->view_scroll_toid == (guint)-1)
		{
		    EditorViewScrollTOCB(e);
		    e->view_scroll_toid = gtk_timeout_add(
			SE_DEF_SCROLL_RATE_MS,
			(GtkFunction)EditorViewScrollTOCB,
			e
		    );
		}
	    }

            status = TRUE;
	}

	return(status);
}

/*
 *	View scroll timeout signal callback.
 */
static gint EditorViewScrollTOCB(gpointer data)
{
	gint xw, yw, width, height;
	gfloat zoom;
	gfloat dxm = 0.0f, dym = 0.0f;
	GdkWindow *window;
	GtkWidget *w;
	editor_struct *e = EDITOR(data);
	if(e == NULL)
	    return(FALSE);

	/* Get view widget and make sure it is mapped and realized. */
	w = e->view_da;
	if((w != NULL) ?
	    (!GTK_WIDGET_MAPPED(w) || !GTK_WIDGET_REALIZED(w))
	    : TRUE
	)
	{
	    e->view_scroll_toid = (guint)-1;
	    return(FALSE);
	}

	/* Get view window. */
	window = w->window;
	if(window == NULL)
	{
            e->view_scroll_toid = (guint)-1;
            return(FALSE);
        }

	/* Get oiinter position in window coordinates and the size of
	 * the window.
	 */
	xw = e->drag_last_x;
	yw = e->drag_last_y;
	gdk_window_get_size(window, &width, &height);
	if((width <= 0) || (height <= 0))
	{
            e->view_scroll_toid = (guint)-1;
            return(FALSE);
        }

	/* Cursor now within the window? */
	if((xw >= 0) && (xw < width) &&
           (yw >= 0) && (yw < height)
	)
	{
            e->view_scroll_toid = (guint)-1;
            return(FALSE);
        }

	/* Get zoom and calculate scroll offset in meters. */
	EditorViewGetViewZoomM(e, &zoom);
	if(zoom <= 0.0f)
	{
	    e->view_scroll_toid = (guint)-1;
	    return(FALSE);
	}
	if(xw < 0)
	    dxm = (gfloat)(0 - xw) * zoom / (gfloat)width;
	if(xw >= width)
	    dxm = (gfloat)(width - xw) * zoom / (gfloat)width;
        if(yw < 0)
            dym = (gfloat)(yw) * zoom / (gfloat)height;
        if(yw >= height)
            dym = (gfloat)(yw - height) * zoom / (gfloat)height;

	/* Convert scroll offset to correct units. */
	switch(EDITOR_UNITS(e))
	{
	  case EDITOR_UNITS_METRIC:
	  case EDITOR_UNITS_METRIC_ALT_FEET:
	    break;
	  case EDITOR_UNITS_ENGLISH:
	    dxm = (gfloat)SFMMetersToFeet(dxm);
	    dym = (gfloat)SFMMetersToFeet(dym);
	    break;
	}

	/* Scroll. */
	if((dxm != 0.0f) || (dym != 0.0f))
	{
	    gfloat xm, ym, zm;
	    GtkAdjustment *adj;
	    GtkWidget *w;

            /* Get previous cursor position in meters. */
            EditorViewGetCursorM(e, NULL, NULL, &zm);
            EditorViewGetPositionWToM(
                e, xw, yw, &xm, &ym
            );

            /* Update position. */
	    e->view_freeze_count++;
            w = e->view_x_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if((adj != NULL) && (dxm != 0.0f))
            {
                adj->value = CLIP(
                    adj->value - dxm,
                    adj->lower, adj->upper - adj->page_size
                );
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }
            w = e->view_y_spin;
            adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
            if((adj != NULL) && (dym != 0.0f))
            {
                adj->value = CLIP(
                    adj->value - dym,
                    adj->lower, adj->upper - adj->page_size
                );
                gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
            }
            e->view_freeze_count--;
            EditorViewAdjValueChangedCB(NULL, e);

	    /* Set new cursor position. */
	    EditorViewSetCursorM(e, xm, ym, zm);
	}

	return(TRUE);
}

/*
 *	View adjustment "value_changed" signal callback.
 */
void EditorViewAdjValueChangedCB(
        GtkAdjustment *adj, gpointer data
)
{
	gint width, height;
	gfloat offset_xm, offset_ym;
	gfloat cursor_xm, cursor_ym, cursor_zm;
	gfloat zoom;
	GdkWindow *window;
	GtkWidget *w;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	w = e->view_da;
        window = (w != NULL) ? w->window : NULL;
	if(window == NULL)
	    return;

	/* Update view cursor position in window coordinates if the view
	 * is not frozen.
	 */
	if(e->view_freeze_count <= 0)
	{
	    gdk_window_get_size(window, &width, &height);
	    EditorViewGetViewPositionM(e, &offset_xm, &offset_ym);
	    EditorViewGetViewZoomM(e, &zoom);
	    EditorViewGetCursorM(e, &cursor_xm, &cursor_ym, &cursor_zm);

	    if(zoom > 0.0f)
	    {
		e->view_cursor_xw = (width / 2) +
		    (gint)((cursor_xm - offset_xm) * width / zoom);
		e->view_cursor_yw = (height / 2) -
		    (gint)((cursor_ym - offset_ym) * height / zoom);
	    }

	    EditorViewDraw(e);
	}
}


/*
 *	Editor view "drag_motion" signal callback.
 */
gboolean EditorViewDragMotionCB(
        GtkWidget *widget, GdkDragContext *dc, gint x, gint y, guint t,
        gpointer data
)
{
        if(dc == NULL)
            return(FALSE);

        if(dc->actions & GDK_ACTION_COPY)
            gdk_drag_status(dc, GDK_ACTION_COPY, t);
        else
            gdk_drag_status(dc, 0, t);

        return(TRUE);
}

/*
 *	Editor view "drag_data_received" signal callback.
 */
void EditorViewDragDataReceivedCB(
        GtkWidget *widget, GdkDragContext *dc, gint x, gint y,
        GtkSelectionData *selection_data, guint info, guint t,
        gpointer data
)
{
	editor_struct *e = EDITOR(data);
        if((selection_data == NULL) || (e == NULL))
            return;

        if((selection_data->data == NULL) ||
           (selection_data->length <= 0)
        )
            return;

        if((info == SE_DND_TYPE_INFO_TEXT_PLAIN) ||
           (info == SE_DND_TYPE_INFO_TEXT_URI_LIST) ||
           (info == SE_DND_TYPE_INFO_STRING)
        )
        {
            const gchar *filename = selection_data->data;


	    EditorSetBusy(e, TRUE);

            /* Seek filename past protocol prefix as needed. */
            if(!strncmp(filename, "file://", 7))
                filename += 7;

            /* Seek past host name and other info until path portion
             * is reached.
             */
            filename = strchr(filename, '/');
            if(filename != NULL)
            {
		/* Check if there are changes and query user for save
		 * as needed, then perform save if user requests it.
		 */
		if(!EditorSceneryDoQuerySave(e))
		{
		    /* Saved succesfully or no save needed/requested,
		     * now load scenery file.
		     */
		    EditorSceneryLoad(e, filename, TRUE);
		    EditorViewDraw(e);
		    EditorUpdateTitle(e);
		    EditorUpdateMenus(e);
		}
	    }

	    EditorSetBusy(e, FALSE);
        }
}

/*
 *	View units changed callback.
 */
void EditorViewUnitsChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
)
{
	gboolean is_alt;
	gint prev_units, new_units;
	GtkAdjustment *adj;
	GtkWidget *w;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	/* Record previous value. */
	prev_units = new_units = e->units;

	/* Get new value. */
	w = e->view_units_combo;
        if(w != NULL)
        {
            GtkCombo *combo = GTK_COMBO(w);
            GtkEntry *entry = GTK_ENTRY(combo->entry);
            const gchar *s1 = gtk_entry_get_text(entry), *s2;
            gint i;
            GList *glist;

            for(i = 0, glist = GUIComboGetList(w);
                glist != NULL;
                i++, glist = glist->next
            )
            {
                s2 = (const gchar *)glist->data;
                if((s1 != NULL) && (s2 != NULL))
                {
                    if(!g_strcasecmp(s1, s2))
			e->units = new_units = i;
                }
            }
        }

	/* No change? */
	if(prev_units == new_units)
	    return;

        e->view_freeze_count++;

#define DO_CONVERT_VALUE				\
{ if(adj != NULL) { 					\
 /* Metric to Metric (alt feet)? */			\
 if((prev_units == EDITOR_UNITS_METRIC) &&		\
    (new_units == EDITOR_UNITS_METRIC_ALT_FEET) &&	\
    is_alt)						\
  adj->value = (gfloat)SFMMetersToFeet(adj->value);	\
 /* Metric to English? */				\
 else if((prev_units == EDITOR_UNITS_METRIC) &&		\
         (new_units == EDITOR_UNITS_ENGLISH))		\
  adj->value = (gfloat)SFMMetersToFeet(adj->value);	\
 /* Metric (alt feet) to Metric? */			\
 else if((prev_units == EDITOR_UNITS_METRIC_ALT_FEET) &&\
         (new_units == EDITOR_UNITS_METRIC) &&		\
         is_alt)					\
  adj->value = (gfloat)SFMFeetToMeters(adj->value);	\
 /* Metric (alt feet) to English? */			\
 else if((prev_units == EDITOR_UNITS_METRIC_ALT_FEET) &&\
         (new_units == EDITOR_UNITS_ENGLISH) &&		\
         !is_alt)					\
  adj->value = (gfloat)SFMMetersToFeet(adj->value);	\
 /* English to Metric? */				\
 else if((prev_units == EDITOR_UNITS_ENGLISH) &&	\
         (new_units == EDITOR_UNITS_METRIC))		\
  adj->value = (gfloat)SFMFeetToMeters(adj->value);	\
 /* English to Metric (alt feet)? */			\
 else if((prev_units == EDITOR_UNITS_ENGLISH) &&	\
         (new_units == EDITOR_UNITS_METRIC_ALT_FEET) &&	\
         !is_alt)					\
  adj->value = (gfloat)SFMFeetToMeters(adj->value);	\
 /* Notify value changed. */				\
 gtk_signal_emit_by_name(				\
  GTK_OBJECT(adj), "value_changed"			\
 );							\
} }

	/* Update displayed units. */
	w = e->view_x_spin;
	adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
	is_alt = FALSE;
	DO_CONVERT_VALUE

        w = e->view_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        DO_CONVERT_VALUE

        w = e->view_zoom_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        DO_CONVERT_VALUE

        w = e->view_grid_spacing_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        DO_CONVERT_VALUE

        w = e->view_cursor_x_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        DO_CONVERT_VALUE

        w = e->view_cursor_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = FALSE;
        DO_CONVERT_VALUE

        w = e->view_cursor_z_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        is_alt = TRUE;
        DO_CONVERT_VALUE

#undef DO_CONVERT_VALUE

        e->view_freeze_count--;
	EditorViewAdjValueChangedCB(NULL, e);
	EditorUpdateMenus(e);
}

/*
 *	View display map callback.
 */
void EditorViewDisplayMapCB(
        GtkWidget *widget, gpointer data
)
{
	gulong flag = EDITOR_VIEW_DISPLAY_MAP;
	editor_struct *e = EDITOR(data);
	if(e == NULL)
	    return;

	if(e->view_display_flags & flag)
	    e->view_display_flags &= ~flag;
	else
	    e->view_display_flags |= flag;

        EditorViewDraw(e);
	EditorUpdateMenus(e);
}

/*
 *      View display object labels callback.
 */
void EditorViewDisplayObjectLabelsCB(
        GtkWidget *widget, gpointer data
)
{
        gulong flag = EDITOR_VIEW_DISPLAY_OBJECT_LABELS;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(e->view_display_flags & flag)
            e->view_display_flags &= ~flag;
        else
            e->view_display_flags |= flag;

        EditorViewDraw(e);
        EditorUpdateMenus(e);
}

/*
 *      View display position label callback.
 */
void EditorViewDisplayPositionLabelCB(
        GtkWidget *widget, gpointer data
)
{
        gulong flag = EDITOR_VIEW_DISPLAY_POSITION_LABEL;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(e->view_display_flags & flag)
            e->view_display_flags &= ~flag;
        else
            e->view_display_flags |= flag;

        EditorViewDraw(e);
        EditorUpdateMenus(e);
}

/*
 *      View display registered locations callback.
 */
void EditorViewDisplayRegisteredLocationsCB(
        GtkWidget *widget, gpointer data
)
{
        gulong flag = EDITOR_VIEW_DISPLAY_REGISTERED_LOCATIONS;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(e->view_display_flags & flag)
            e->view_display_flags &= ~flag;
        else
            e->view_display_flags |= flag;

	EditorViewDraw(e);
        EditorUpdateMenus(e);
}


/*
 *      Object list "key_press_event" signal callback.
 */
gint EditorObjListKeyEventCB(
        GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
        gint status = FALSE;
        gboolean key_press;
        editor_struct *e = EDITOR(data);
        if((key == NULL) || (e == NULL))
            return(status);

        key_press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;

#define DO_STOP_KEY_SIGNAL      \
{ if(widget != NULL) \
 gtk_signal_emit_stop_by_name( \
  GTK_OBJECT(widget), \
  key_press ? "key_press_event" : "key_release_event" \
 ); \
}

        switch(key->keyval)
        {
          case GDK_Return:
          case GDK_KP_Enter:
          case GDK_ISO_Enter:
          case GDK_3270_Enter:

            status = TRUE;
            break;

          case GDK_Insert:
            if(key_press)
            {
		if(key->state & GDK_CONTROL_MASK)
		    EditorCopyCB(NULL, e);
		else if(key->state & GDK_SHIFT_MASK)
		    EditorPasteCB(NULL, e);
		else if(key->state == 0x00000000)
		    EditorNewObjectCB(NULL, e);
            }
            status = TRUE;
            break;

	  case GDK_Delete:
	    if(key_press)
	    {
		EditorDeleteObjectCB(NULL, e);
	    }
	    status = TRUE;
	    break;
        }

#undef DO_STOP_KEY_SIGNAL

        return(status);
}

/*
 *	Object list "button_press_event" signal callback.
 */
gint EditorObjListButtonEventCB(
        GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
        gint status = FALSE;
	gint row, column;
	GtkCList *clist = (GtkCList *)widget;
        editor_struct *e = EDITOR(data);
        if((clist == NULL) || (button == NULL) || (e == NULL))
            return(status);

        if(clist->clist_window != ((GdkEventAny *)button)->window)
            return(status);

        switch((gint)button->type)
        {
	  case GDK_BUTTON_PRESS:
	    e->drag_button = button->button;
	    if(!gtk_clist_get_selection_info(
		clist, (gint)button->x, (gint)button->y,
		&row, &column
	    ))
	    {
		row = 0;
		column = 0;
	    }
	    switch(button->button)
	    {
              case 3:
		if(!(button->state & GDK_CONTROL_MASK) &&
                   !(button->state & GDK_SHIFT_MASK)
		)
		    gtk_clist_unselect_all(clist);
		gtk_clist_select_row(clist, row, column);
                if(e->obj_menu != NULL)
                {
                    GtkMenu *menu = GTK_MENU(e->obj_menu);
                    gtk_menu_popup(
                        menu, NULL, NULL,
                        NULL, NULL,
                        button->button, button->time
                    );
                }
		status = TRUE;
		break;
	    }
            break;

          case GDK_BUTTON_RELEASE:
            e->drag_button = 0;
            if(gdk_pointer_is_grabbed())
            {
                gdk_pointer_ungrab(button->time);
                status = TRUE;
            }
            break;
        }

        return(status);
}

/*
 *	Object model file browse callback.
 */
void EditorObjModelBrowseCB(gpointer widget, gpointer data)
{
        gboolean status;
        const gchar *s, *cwd;
        gchar *filename, *filename_short;
        fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
        gint total_ftypes = 0;
        gchar **path_rtn = NULL;
        gint total_path_rtns = 0;
	GtkWidget *w;
        se_core_struct *core_ptr;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(FileBrowserIsQuery())
            return;

        core_ptr = SE_CORE(e->core_ptr);

        EditorSetBusy(e, TRUE);

        /* Get current working directory. */
        cwd = (core_ptr != NULL) ? core_ptr->cwd : NULL;
        if(cwd == NULL)
#ifdef __MSW__
            cwd = "\\";
#else
            cwd = "/";
#endif

        /* Create file types list. */
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            ".v3d .3d", "Vertex 3D Object files"
        );
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            "*.*", "All files"
        );

	/* Get startup path. */
	w = e->obj_model_entry;
#ifdef __MSW__
        filename = STRDUP((w != NULL) ?
            gtk_entry_get_text(GTK_ENTRY(w)) : cwd
        );
        if(!ISPATHABSOLUTE(filename))
        {
            g_free(filename);
            filename = STRDUP(cwd);
        }
#else
	filename = STRDUP((w != NULL) ?
	    gtk_entry_get_text(GTK_ENTRY(w)) : SE_DEF_SAR_DATA_DIR
	);
	if(!ISPATHABSOLUTE(filename))
	{
	    g_free(filename);
	    filename = STRDUP(SE_DEF_SAR_DATA_DIR);
	}
#endif

        /* Query user for scenery file. */
        FileBrowserSetTransientFor(EDITOR_TOPLEVEL(e));
        status = FileBrowserGetResponse(
            "Select Model File",
            "Select", "Cancel",
	    filename,
            ftype, total_ftypes,
            &path_rtn, &total_path_rtns,
            &ftype_rtn
        );
        FileBrowserSetTransientFor(NULL);
        FileBrowserDeleteTypeList(ftype, total_ftypes);
	g_free(filename);

        /* Get filename if there was a response. */
        s = (status && (total_path_rtns > 0)) ?
            path_rtn[0] : NULL;
        filename = STRDUP(s);
        if(filename == NULL)
        {
            EditorSetBusy(e, FALSE);
            return;
        }

        /* Get shortened filename if the filename has a prefix of the
         * data directory.
         */
#ifdef __MSW__
        if(strpfx(filename, cwd))
	{
            s = filename + strlen(cwd);
	    while(*s == DIR_DELIMINATOR)
		s++;
	}
#else
        if(strpfx(filename, SE_DEF_SAR_DATA_DIR))
	{
	    s = filename + strlen(SE_DEF_SAR_DATA_DIR);
	    while(*s == DIR_DELIMINATOR)
                s++;
	}
#endif
        else
	{
	    s = filename;
	}
	filename_short = STRDUP(s);

	/* Set new model file name. */
	if(w != NULL)
	    gtk_entry_set_text(GTK_ENTRY(w), filename_short);

	g_free(filename_short);
        g_free(filename);

        EditorSetBusy(e, FALSE);

        EditorViewDraw(e);
}

/*
 *	Object type changed callback.
 */
void EditorObjTypeChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
)
{
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(e->obj_freeze_count <= 0)
	{
	    /* Need to first apply values then get values because
	     * certain widgets need to be mapped or unmapped depending
	     * on the type change.
	     */
            EditorObjListDoApplyValues(e, NULL);
	    if(e->obj_clist != NULL)
	    {
		GtkCList *clist = (GtkCList *)e->obj_clist;
		GList *glist = clist->selection_end;
		gint row = (glist != NULL) ? (gint)glist->data : -1;
		EditorObjListDoGetValues(
		    e,
		    row,
		    (row > -1) ? gtk_clist_get_row_data(clist, row) : NULL
		);
	    }
	}
}

/*
 *	Object polygon offset GtkToggleButton "toggled" signal callback.
 */
void EditorObjPolygonOffsetToggledCB(gpointer widget, gpointer data)
{
	gboolean sensitive;
	GtkWidget *w;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

#define DO_SET_SENSITIVE	\
{ if(w != NULL) { gtk_widget_set_sensitive(w, sensitive); } }

	w = e->obj_polygon_offset_tb;
	sensitive = (w != NULL) ? GTK_TOGGLE_BUTTON(w)->active : TRUE;
	w = e->obj_polygon_offset_reverse_tb;
	DO_SET_SENSITIVE
	w = e->obj_polygon_offset_write_depth_tb;
	DO_SET_SENSITIVE
}


/*
 *	Object size shape changed callback.
 */
void EditorObjSizeShapeChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
)
{
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(e->obj_freeze_count <= 0)
            EditorObjListDoApplyValues(e, NULL);
}

/*
 *	Object adjustment "value_changed" signal callback.
 */
void EditorObjAdjValueChangedCB(
        GtkAdjustment *adj, gpointer data
)
{
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	if(e->obj_freeze_count <= 0)
	    EditorObjListDoApplyValues(e, NULL);
}

/*
 *      Object reference adjustment "value_changed" signal callback.
 */
void EditorObjReferenceAdjValueChangedCB(
        GtkAdjustment *adj, gpointer data
)
{
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(e->obj_freeze_count <= 0)
        {
            EditorObjListDoApplyValues(e, NULL);
            EditorObjListRealizeAll(e, TRUE);
	    EditorViewDraw(e);
        }
}

/*
 *	Object reference change callback.
 */
void EditorObjReferenceChangeCB(GtkWidget *widget, gpointer data)
{
	gchar **strv;
	gint strc;
	GtkWidget *toplevel;
	editor_obj_struct *o;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	if(PDialogIsQuery())
	    return;

	toplevel = EDITOR_TOPLEVEL(e);
	o = EDITOR_OBJ(EditorObjListGetObj(e, e->obj_selected));
	if(o == NULL)
	    return;

	EditorSetBusy(e, TRUE);

	PDialogDeleteAllPrompts();
	PDialogAddPrompt(NULL, "Name:", o->ref_obj_name);
	PDialogSetSize(350, -1);
	PDialogSetTransientFor(toplevel);
	strv = PDialogGetResponse(
	    "Reference Object",
"Enter the name of the reference object that you want\n\
this object to take relative positions with (or leave\n\
blank to specify no reference object).\n\
\n\
Special names:\n\
player                  Player object\n\
camera                  Camera\n\
ear                     Ear\n\
\n",
	    NULL,
	    PDIALOG_ICON_QUESTION,
	    "Set", "Cancel",
	    PDIALOG_BTNFLAG_SUBMIT | PDIALOG_BTNFLAG_CANCEL,
	    PDIALOG_BTNFLAG_SUBMIT,
	    &strc
	);
        PDialogSetTransientFor(NULL);
	if(strv != NULL)
	{
	    GtkWidget *w;
	    const gchar *name = (strc > 0) ? strv[0] : NULL;
	    if((name != NULL) ? (*name != '\0') : FALSE)
	    {
		gint on = EditorObjListMatchByName(e, name);
		if(on > -1)
		{
		    /* Matched object. */
                    g_free(o->ref_obj_name);
                    o->ref_obj_name = STRDUP(name);
		    o->ref_obj_num = on;
		    w = e->obj_ref_name_entry;
		    if(w != NULL)
			gtk_entry_set_text(GTK_ENTRY(w), name);
		}
		else
		{
		    /* Object could not be found, check if it is a
		     * special name.
		     */
		    if(EditorIsObjectNameSpecial(name))
		    {
			/* This is a special name, so only set the name
			 * but not update the reference object.
			 */
			g_free(o->ref_obj_name);
                        o->ref_obj_name = STRDUP(name);
                        o->ref_obj_num = -1;
                        w = e->obj_ref_name_entry;
                        if(w != NULL)
                            gtk_entry_set_text(GTK_ENTRY(w), name);
		    }
		    else
		    {
			/* Reference object could not be found, query user
			 * if they still want to use the value for the
			 * reference object name on this object.
			 */
			gint status;
			gchar *buf = g_strdup_printf(
"Unable to find object \"%s\",\nUse this value anyway?",
			    name
			);
			CDialogSetTransientFor(toplevel);
			status = CDialogGetResponse(
			    "Object Not Found",
			    buf,
"The given object name could not be found.  If you\n\
still wish to use that as the reference object name\n\
for this object then click on \"yes\", otherwise\n\
click on \"no\".",
			    CDIALOG_ICON_WARNING,
			    CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
			    CDIALOG_BTNFLAG_NO
			);
			CDialogSetTransientFor(NULL);
			g_free(buf);
			if(status == CDIALOG_RESPONSE_YES)
			{
                            g_free(o->ref_obj_name);
                            o->ref_obj_name = STRDUP(name);
                            o->ref_obj_num = -1;
                            w = e->obj_ref_name_entry;
                            if(w != NULL)
                                gtk_entry_set_text(GTK_ENTRY(w), name);
			}
		    }
		}
	    }
	    else
	    {
		/* Got blank name, so remove reference object name on
		 * this object.
		 */
		g_free(o->ref_obj_name);
		o->ref_obj_name = NULL;
		o->ref_obj_num = -1;
                w = e->obj_ref_name_entry;
                if(w != NULL)
                    gtk_entry_set_text(GTK_ENTRY(w), "");
	    }
	}

        EditorSetBusy(e, FALSE);
}

/*
 *	Preset model type changed callback.
 */
void EditorObjPresetModelTypeChangedCB(
        GtkWidget *widget, gpointer data, GList *glist
)
{
	gint i;
	gchar *tip;
	const gchar *type, *value_desc = NULL;
	GtkWidget *w;
	GtkEntry *entry;
	gchar *type_names[] = SE_SAR_PRESET_MODEL_NAMES;
	gchar *value_descs[] = SE_SAR_PRESET_MODEL_VALUES_DESC;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	/* Get current type name. */
	w = e->obj_preset_model_type_combo;
	entry = (GtkEntry *)((w != NULL) ?
	    GTK_COMBO(w)->entry : NULL
	);
	type = (entry != NULL) ? gtk_entry_get_text(entry) : NULL;
	if(type == NULL)
	    return;

	/* Check if the preset model type name is in the list of known
	 * types.
	 */
	for(i = 0; type_names[i] != NULL; i++)
	{
	    if(!g_strcasecmp(type_names[i], type))
		break;
	}
	/* Found known preset model type? */
	if(type_names[i] != NULL)
	{
	    /* Get matching values description, since we know that the
	     * index i reffers to a valid preset model type name then
	     * i corresponds to a valid index in the value_descs list.
	     */
	    value_desc = value_descs[i];
	}
	if(value_desc == NULL)
	    tip = STRDUP("*Values format unknown*");
	else
	    tip = g_strdup_printf(
		"\"%s\" values: %s",
		type, value_desc
	    );

	w = e->obj_preset_model_values_entry;
	if(w != NULL)
	{
	    GUISetWidgetTip(w, tip);
	    GUIShowTipsNow(w);
	}

	g_free(tip);
}


/*
 *	Toolbar item callback.
 */
void EditorToolBarCB(gpointer data)
{
	GtkWidget *toplevel;
	editor_struct *e;
	editor_opid_struct *opid = EDITOR_OPID(data);
	if(opid == NULL)
	    return;

	e = EDITOR(opid->editor);
	if(e == NULL)
	    return;

	toplevel = EDITOR_TOPLEVEL(e);

	if(opid->func_cb != NULL)
	    opid->func_cb(toplevel, e);
}

/*
 *	Toolbar item enter callback.
 */
void EditorToolBarEnterCB(gpointer data)
{
        editor_struct *e;
        editor_opid_struct *opid = EDITOR_OPID(data);
        if(opid == NULL)
            return;

        e = EDITOR(opid->editor);
        if(e == NULL)
            return;

	StatusBarMessage(e->status_bar, opid->tooltip, FALSE);
}

/*
 *	Toolbar item leave callback.
 */
void EditorToolBarLeaveCB(gpointer data)
{
        editor_struct *e;
        editor_opid_struct *opid = EDITOR_OPID(data);
        if(opid == NULL)
            return;

        e = EDITOR(opid->editor);
        if(e == NULL)
            return;

        StatusBarMessage(e->status_bar, NULL, FALSE);
}

/*
 *	New scenery callback.
 */
void EditorNewCB(GtkWidget *widget, gpointer data)
{
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        EditorSetBusy(e, TRUE);

        /* Check if there are changes and query user for save as needed,
         * then perform save if user requests it.
         */
        if(EditorSceneryDoQuerySave(e))
	{
	    EditorSetBusy(e, FALSE);
            return;
	}

	EditorSceneryUnload(e);

        EditorSetBusy(e, FALSE);

        EditorViewDraw(e);
        EditorUpdateMenus(e);
}

/*
 *	Open scenery callback.
 */
void EditorOpenCB(GtkWidget *widget, gpointer data)
{
	gboolean status;
	const gchar *s, *cwd;
	gchar *filename;
        fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
        gint total_ftypes = 0;
        gchar **path_rtn = NULL;
        gint total_path_rtns = 0;
        se_core_struct *core_ptr;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(FileBrowserIsQuery())
            return;

        core_ptr = SE_CORE(e->core_ptr);

        EditorSetBusy(e, TRUE);

        /* Apply changes to currently selected object. */
        EditorObjListDoApplyValues(e, NULL);

        /* Get current working directory. */
        cwd = (core_ptr != NULL) ? core_ptr->cwd : NULL;
        if(cwd == NULL)
#ifdef __MSW__
            cwd = "\\";
#else
            cwd = "/";
#endif

        /* Check if there are changes and query user for save as needed,
         * then perform save if user requests it.
         */
        if(EditorSceneryDoQuerySave(e))
        {
            EditorSetBusy(e, FALSE);
            return;
        }

	/* Create file types list. */
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            ".scn", "Search and Rescue Scenery files"
        );
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            "*.*", "All files"
        );

	/* Get initial path. */
	filename = STRDUP((e->scn_filename != NULL) ?
	    e->scn_filename : SE_DEF_SAR_DATA_SCENERY_DIR
	);

        /* Query user for scenery file. */
        FileBrowserSetTransientFor(EDITOR_TOPLEVEL(e));
        status = FileBrowserGetResponse(
            "Open Scenery File",
            "Open", "Cancel",
            filename,
            ftype, total_ftypes,
            &path_rtn, &total_path_rtns,
            &ftype_rtn
        );
        FileBrowserSetTransientFor(NULL);
        FileBrowserDeleteTypeList(ftype, total_ftypes);
	g_free(filename);

        /* Get filename if there was a response. */
        s = (status && (total_path_rtns > 0)) ?
            path_rtn[0] : NULL;
        filename = STRDUP(s);
        if(filename == NULL)
        {
            EditorSetBusy(e, FALSE);
            return;
        }

	/* Load scenery file. */
	EditorSceneryLoad(e, filename, TRUE);
	g_free(filename);

        EditorSetBusy(e, FALSE);

	EditorViewDraw(e);
	EditorUpdateTitle(e);
	EditorUpdateMenus(e);
}

/*
 *	Save callback.
 */
void EditorSaveCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	toplevel = EDITOR_TOPLEVEL(e);

	/* Untitled scenery? */
	if(e->scn_filename == NULL)
	{
	    /* Need to perform a saveas. */
	    EditorSaveAsCB(NULL, e);
	}
	else
	{
	    /* Save scenery file. */
            EditorSetBusy(e, TRUE);
	    EditorScenerySave(e, e->scn_filename, TRUE);
            EditorUpdateTitle(e);
            EditorUpdateMenus(e);
            EditorSetBusy(e, FALSE);
	}
}

/*
 *	Save as callback.
 */
void EditorSaveAsCB(GtkWidget *widget, gpointer data)
{
        gboolean status;
        const gchar *s, *cwd;
        gchar *filename;
        fb_type_struct **ftype = NULL, *ftype_rtn = NULL;
        gint total_ftypes = 0;
        gchar **path_rtn = NULL;
        gint total_path_rtns = 0;
	struct stat stat_buf;
        se_core_struct *core_ptr;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        if(FileBrowserIsQuery())
            return;

        core_ptr = SE_CORE(e->core_ptr);

        EditorSetBusy(e, TRUE);

        /* Get current working directory. */
        cwd = (core_ptr != NULL) ? core_ptr->cwd : NULL;
        if(cwd == NULL)
#ifdef __MSW__
            cwd = "\\";
#else
            cwd = "/";
#endif

        /* Create file types list. */
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            ".scn", "Search and Rescue Scenery files"
        );
        FileBrowserTypeListNew(
            &ftype, &total_ftypes,
            "*.*", "All files"
        );

	/* Get initial path. */
	filename = STRDUP((e->scn_filename != NULL) ?
	    e->scn_filename : SE_DEF_SAR_DATA_SCENERY_DIR
	);

        /* Query user for scenery file. */
        FileBrowserSetTransientFor(EDITOR_TOPLEVEL(e));
        status = FileBrowserGetResponse(
            "Save Scenery File",
            "Save", "Cancel",
	    filename,
            ftype, total_ftypes,
            &path_rtn, &total_path_rtns,
            &ftype_rtn
        );
        FileBrowserSetTransientFor(NULL);
        FileBrowserDeleteTypeList(ftype, total_ftypes);
	g_free(filename);

        /* Get filename if there was a response. */
        s = (status && (total_path_rtns > 0)) ?
            path_rtn[0] : NULL;
        filename = STRDUP(s);
        if(filename == NULL)
        {
            EditorSetBusy(e, FALSE);
            return;
        }

	/* Check if file exists. */
	if(!stat(filename, &stat_buf))
	{
	    gint status;
	    gchar *buf = g_strdup_printf(
		"Overwrite existing file:\n\n    %s",
		filename
	    );
            CDialogSetTransientFor(EDITOR_TOPLEVEL(e));
            status = CDialogGetResponse(
                "Confirm Overwrite",
		buf,
		NULL,
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                CDIALOG_BTNFLAG_NO
            );
            CDialogSetTransientFor(NULL);
	    g_free(buf);
	    if(status != CDIALOG_RESPONSE_YES)
	    {
		g_free(filename);
                EditorSetBusy(e, FALSE);
                return;
	    }
	}

	/* Update scenery filename. */
	g_free(e->scn_filename);
	e->scn_filename = STRDUP(filename);

        /* Save scenery file. */
        EditorScenerySave(e, filename, TRUE);
        g_free(filename);

        EditorSetBusy(e, FALSE);

        EditorUpdateTitle(e);
        EditorUpdateMenus(e);
}

/*
 *	Close callback.
 */
void EditorCloseCB(GtkWidget *widget, gpointer data)
{
	editor_struct *e = EDITOR(data);
	if(e == NULL)
	    return;

        EditorSetBusy(e, TRUE);

	/* Check if the editor is still processing. */
	if(EDITOR_PROCESSING(e))
	{
            CDialogSetTransientFor(EDITOR_TOPLEVEL(e));
	    CDialogGetResponse(
		"Operation In Progress",
"There is an operation currently in progress, please\n\
wait for it to be completed before using this function.\n\
To stop the current operation click on \"Stop\"",
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    EditorSetBusy(e, FALSE);
	    return;
	}

        /* Check if there are changes and query user for save as needed,
         * then perform save if user requests it.
         */
        if(EditorSceneryDoQuerySave(e))
	{
	    EditorSetBusy(e, FALSE);
            return;
	}

        EditorSetBusy(e, FALSE);

	EditorSavePosition(e);
	EditorUnmap(e);
}

/*
 *	Cut callback.
 */
void EditorCutCB(GtkWidget *widget, gpointer data)
{
	gchar *buf;
	gint objects_cut = 0;
        GtkCList *clist;
        se_core_struct *core_ptr;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        core_ptr = SE_CORE(e->core_ptr);
        if(core_ptr == NULL)
            return;

        clist = (GtkCList *)e->obj_clist;
        if(clist == NULL)
            return;

	if(clist->selection == NULL)
	    return;

        EditorSetBusy(e, TRUE);

	/* Copy selected objects. */
	EditorCopyCB(widget, data);

        StatusBarMessage(e->status_bar, "Deleting objects...", TRUE);

	/* Delete selected objects. */
	while(clist->selection != NULL)
	{
	    gtk_clist_remove(clist, (gint)clist->selection->data);
	    objects_cut++;
            StatusBarProgress(e->status_bar, -1.0f, TRUE);
	}

        buf = g_strdup_printf(
            "%i object%s cut",
            objects_cut,
            (objects_cut != 1) ? "s" : ""
        );
        StatusBarMessage(e->status_bar, buf, TRUE);
        g_free(buf);
        StatusBarProgress(e->status_bar, 0.0f, TRUE);

	if(objects_cut > 0)
	{
	    if(!EDITOR_HAS_CHANGES(e))
	    {
		e->has_changes = TRUE;
		EditorUpdateTitle(e);
	    }
            EditorObjListRealizeAll(e, TRUE);
            EditorViewDraw(e);
            EditorUpdateMenus(e);
	}

        EditorSetBusy(e, FALSE);
}

/*
 *	Copy callback.
 */
void EditorCopyCB(GtkWidget *widget, gpointer data)
{
	gchar *buf;
	gint objects_coppied = 0;
	GList *glist;
	GtkCList *clist;
	se_dde_struct *dde;
	se_core_struct *core_ptr;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	core_ptr = SE_CORE(e->core_ptr);
	if(core_ptr == NULL)
	    return;

	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return;

	if(clist->selection == NULL)
	    return;

        EditorSetBusy(e, TRUE);

	StatusBarMessage(e->status_bar, "Copying objects...", TRUE);

	/* Delete old DDE and create a new DDE. */
	SEDDEDelete(core_ptr->dde);
	core_ptr->dde = dde = SEDDENew(core_ptr, 0);
	/* Iterate through selected objects, adding them to the dde. */
	for(glist = clist->selection; glist != NULL; glist = glist->next)
	{
	    gint row = (gint)glist->data;
	    const editor_obj_struct *obj_src = EDITOR_OBJ(
                gtk_clist_get_row_data(clist, row)
            );
            editor_obj_struct *obj_tar = EditorObjDup(obj_src);
	    SEDDEObjectAdd(dde, obj_tar);
	    objects_coppied++;

	    StatusBarProgress(e->status_bar, -1.0f, TRUE);
	}

        buf = g_strdup_printf(
            "%i object%s coppied",
            objects_coppied,
            (objects_coppied != 1) ? "s" : ""
        );
        StatusBarMessage(e->status_bar, buf, TRUE);
        g_free(buf);
        StatusBarProgress(e->status_bar, 0.0f, TRUE);

        EditorSetBusy(e, FALSE);
}

/*
 *	Paste callback.
 */
void EditorPasteCB(GtkWidget *widget, gpointer data)
{
	gint i;
	gint objects_pasted = 0;
	gint insert_row_tar;
	gchar *buf;
	GList *glist;
        GtkCList *clist_tar, *clist_src;
        se_dde_struct *dde;
        se_core_struct *core_ptr;
        editor_struct *e_tar = EDITOR(data);
	editor_struct *e_src;
        if(e_tar == NULL)
            return;

        core_ptr = SE_CORE(e_tar->core_ptr);
        if(core_ptr == NULL)
            return;

        clist_tar = (GtkCList *)e_tar->obj_clist;
        if(clist_tar == NULL)
            return;

	glist = clist_tar->selection_end;
	insert_row_tar = (glist != NULL) ? (gint)glist->data : -1;

	/* Get DDE. */
	dde = core_ptr->dde;
	if(dde == NULL)
	    return;

	/* Use source editor as the target (this) editor for now. */
	e_src = e_tar;
	if(e_src == NULL)
	    return;

	/* Get source clist. */
        clist_src = (GtkCList *)e_src->obj_clist;
        if(clist_src == NULL)
            return;

        EditorSetBusy(e_tar, TRUE);

        StatusBarMessage(e_tar->status_bar, "Pasting objects...", TRUE);

	/* Iterate through dde objects. */
	for(i = 0; i < dde->total_objects; i++)
	{
	    editor_obj_struct *obj = EditorObjDup(
		EDITOR_OBJ(dde->object[i])
	    );
	    if(obj != NULL)
	    {
		/* Insert new object to list. */
		gint new_row = EditorObjListInsertRow(
		    e_tar, insert_row_tar, obj
		);
		/* Increment insert row on success. */
		if((new_row > -1) && (insert_row_tar > -1))
		{
		    insert_row_tar++;
		    objects_pasted++;

		    StatusBarProgress(
			e_tar->status_bar,
			(gfloat)(i + 1) / (gfloat)dde->total_objects,
			TRUE
		    );
		}
	    }
	}

	buf = g_strdup_printf(
	    "%i object%s pasted",
	    objects_pasted,
	    (objects_pasted != 1) ? "s" : ""
	);
	StatusBarMessage(e_tar->status_bar, buf, TRUE);
	g_free(buf);
        StatusBarProgress(e_tar->status_bar, 0.0f, TRUE);

	if(objects_pasted > 0)
	{
	    if(!EDITOR_HAS_CHANGES(e_tar))
	    {
		e_tar->has_changes = TRUE;
		EditorUpdateTitle(e_tar);
	    }
	    EditorObjListRealizeAll(e_tar, TRUE);
	    EditorViewDraw(e_tar);
	    EditorUpdateMenus(e_tar);
	}

        EditorSetBusy(e_tar, FALSE);
}

/*
 *	New object callback.
 */
void EditorNewObjectCB(GtkWidget *widget, gpointer data)
{
	gchar *buf;
	gint row, new_row;
	gfloat xm, ym, zm;
	GList *glist;
	GtkCList *clist;
        editor_obj_struct *obj;
	sar_position_struct *pos;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        clist = (GtkCList *)e->obj_clist;
        if(clist == NULL)
            return;

	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;

        EditorSetBusy(e, TRUE);

	/* Unselect previous objects, this will apply the values of
	 * the previosly selected object (if any).
	 */
	gtk_clist_unselect_all(clist);

	/* Create a new object. */
	obj = EditorObjNew();
	if(obj == NULL)
	{
	    EditorSetBusy(e, FALSE);
	    return;
	}

	/* Get cursor position in meters. */
	EditorViewGetCursorM(e, &xm, &ym, &zm);

	/* Move object to cursor. */
	pos = &obj->pos;
	pos->x = xm;
        pos->y = ym;
        pos->z = zm;

	/* Insert object to list. */
	new_row = EditorObjListInsertRow(e, row, obj);
	obj = NULL;
	if(new_row > -1)
	    gtk_clist_select_row(clist, new_row, 0);

	if(new_row > -1)
	    buf = g_strdup_printf("Created new object %i", new_row);
	else
	    buf = STRDUP("Create new object failed");
        StatusBarMessage(e->status_bar, buf, TRUE);
        g_free(buf);

	if(!EDITOR_HAS_CHANGES(e))
	{
	    e->has_changes = TRUE;
	    EditorUpdateTitle(e);
	}
	EditorObjListRealizeAll(e, TRUE);
	EditorViewDraw(e);
	EditorUpdateMenus(e);

        EditorSetBusy(e, FALSE);
}

/*
 *	Goto selected callback.
 */
void EditorGoToSelectedCB(GtkWidget *widget, gpointer data)
{
	gint row;
	gfloat xm, ym;
	GList *glist;
	GtkAdjustment *adj;
	GtkWidget *w;
        GtkCList *clist;
	editor_obj_struct *obj;
	const sar_position_struct *pos;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	clist = (GtkCList *)e->obj_clist;
	if(clist == NULL)
	    return;

	/* Get last selected row. */
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	if(row < 0)
	    return;

	/* Get selected object. */
	obj = EDITOR_OBJ(gtk_clist_get_row_data(clist, row));
	if(obj == NULL)
	    return;

	/* Get position of object in meters. */
	pos = &obj->pos;
	xm = (gfloat)pos->x;
	ym = (gfloat)pos->y;

        /* Convert to displayed units. */
        switch(EDITOR_UNITS(e))
        {
          case EDITOR_UNITS_METRIC:
          case EDITOR_UNITS_METRIC_ALT_FEET:
            break;
          case EDITOR_UNITS_ENGLISH:
            xm = (gfloat)SFMMetersToFeet(xm);
            ym = (gfloat)SFMMetersToFeet(ym);
            break;
        }

	e->view_freeze_count++;

        /* Update position. */
        w = e->view_x_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = xm;
            if(adj->value > (adj->upper - adj->page_size))
                adj->value = adj->upper - adj->page_size;
            if(adj->value < adj->lower)
                adj->value = adj->lower;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

        w = e->view_y_spin;
        adj = (w != NULL) ? GTK_SPIN_BUTTON(w)->adjustment : NULL;
        if(adj != NULL)
        {
            adj->value = ym;
            if(adj->value > (adj->upper - adj->page_size))
                adj->value = adj->upper - adj->page_size;
            if(adj->value < adj->lower)
                adj->value = adj->lower;
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");
        }

        e->view_freeze_count--;
        EditorViewAdjValueChangedCB(NULL, e); 
}

/*
 *	Move object to cursor callback.
 */
void EditorMoveToCursorCB(GtkWidget *widget, gpointer data)
{
	gchar *buf;
	gint objects_selected;
        gint objects_moved = 0;
        gfloat xm, ym, zm;
        GList *glist;
        GtkCList *clist;
	sar_position_struct *pos;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        clist = (GtkCList *)e->obj_clist;
        if(clist == NULL)
            return;

        /* Get number of objects selected. */
        objects_selected = (clist->selection != NULL) ?
            g_list_length(clist->selection) : 0;
	if(objects_selected <= 0)
	    return;

        EditorSetBusy(e, TRUE);

        /* Get cursor position in meters. */
        EditorViewGetCursorM(e, &xm, &ym, &zm);

	/* Iterate through all selected objects. */
	for(glist = clist->selection; glist != NULL; glist = glist->next)
	{
	    editor_obj_struct *obj = EDITOR_OBJ(
		gtk_clist_get_row_data(clist, (gint)glist->data)
	    );
	    if(obj == NULL)
		continue;

	    /* If only one object is selected then we need to apply
	     * its object values first.
	     */
	    if(objects_selected == 1)
		EditorObjListDoApplyValues(e, obj);

	    /* Move object to cursor. */
	    pos = &obj->pos;
            pos->x = xm;
            pos->y = ym;
            pos->z = zm;

	    objects_moved++;

            /* Refetch values for this object if it was the only one
	     * selected.
             */
            if(objects_selected == 1)
                EditorObjListDoGetValues(e, (gint)glist->data, obj);
	}

	buf = g_strdup_printf(
	    "%i object%s moved",
	    objects_moved,
	    (objects_moved != 1) ? "s" : ""
	);
        StatusBarMessage(e->status_bar, buf, TRUE);
        g_free(buf);

        if(!EDITOR_HAS_CHANGES(e))
        {
            e->has_changes = TRUE;
            EditorUpdateTitle(e);
        }
	EditorObjListRealizeAll(e, TRUE);	/* Incase obj(s) references this obj. */
        EditorViewDraw(e);
        EditorUpdateMenus(e);

        EditorSetBusy(e, FALSE);
}

/*
 *	Delete object callback.
 */
void EditorDeleteObjectCB(GtkWidget *widget, gpointer data)
{
        gchar *buf;
	gint status;
        gint objects_deleted = 0;
	gint total_selected;
	GList *glist;
	GtkWidget *toplevel;
        GtkCList *clist;
	editor_obj_struct *obj;
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

	toplevel = EDITOR_TOPLEVEL(e);
        clist = (GtkCList *)e->obj_clist;
        if(clist == NULL)
            return;

	/* Get last selected object and number of selected objects. */
	glist = clist->selection_end;
	if(glist == NULL)
	    return;
	total_selected = g_list_length(clist->selection);
	obj = EDITOR_OBJ((total_selected == 1) ?
	    gtk_clist_get_row_data(clist, (gint)glist->data) : NULL
	);

        EditorSetBusy(e, TRUE);

	/* Get user confirmation. */
	if(obj != NULL)
	{
	    if(obj->name != NULL)
		buf = g_strdup_printf(
		    "Delete object \"%s\"?",
		    obj->name
		);
	    else
		buf = g_strdup_printf(
		    "Delete object #%i?",
		    (glist != NULL) ? (gint)glist->data : -1
		);
	}
	else
	{
	    buf = g_strdup_printf(
		"Delete %i object%s?",
		total_selected,
		(total_selected != 1) ? "s" : ""
	    );
	}
        CDialogSetTransientFor(toplevel);
	status = CDialogGetResponse(
	    "Confirm Delete Objects",
	    buf,
	    NULL,
            CDIALOG_ICON_WARNING,
            CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
	    CDIALOG_BTNFLAG_NO
	);
	CDialogSetTransientFor(NULL);
	g_free(buf);
	switch(status)
	{
	  case CDIALOG_RESPONSE_YES_TO_ALL:
          case CDIALOG_RESPONSE_YES:
	    break;

	  default:
	    EditorSetBusy(e, FALSE);
	    return;
	    break;
	}

	/* Delete all selected objects. */
	while(clist->selection != NULL)
	{
	    gtk_clist_remove(clist, (gint)clist->selection->data);
	    objects_deleted++;
	}

	buf = g_strdup_printf(
	    "%i object%s deleted",
	    objects_deleted,
	    (objects_deleted != 1) ? "s" : ""
	);
        StatusBarMessage(e->status_bar, buf, TRUE);
        g_free(buf);

        if(!EDITOR_HAS_CHANGES(e))
        {
            e->has_changes = TRUE;
            EditorUpdateTitle(e);
        }
        EditorObjListRealizeAll(e, TRUE);
        EditorViewDraw(e);
        EditorUpdateMenus(e);

        EditorSetBusy(e, FALSE);
}

/*
 *	Switch to header page callback.
 */
void EditorViewHeaderCB(GtkWidget *widget, gpointer data)
{
	GtkNotebook *notebook = (GtkNotebook *)EDITOR_NOTEBOOK(data);
	if(notebook != NULL)
	    gtk_notebook_set_page(notebook, 0);
}

/*
 *	Switch to scenery page callback.
 */
void EditorViewSceneryCB(GtkWidget *widget, gpointer data)
{
        GtkNotebook *notebook = (GtkNotebook *)EDITOR_NOTEBOOK(data);
        if(notebook != NULL)
            gtk_notebook_set_page(notebook, 1);
}

/*
 *	Refresh callback.
 */
void EditorRefreshCB(GtkWidget *widget, gpointer data)
{
        editor_struct *e = EDITOR(data);
        if(e == NULL)
            return;

        EditorSetBusy(e, TRUE);

        EditorObjListRealizeAll(e, TRUE);
        EditorViewDraw(e);
        EditorUpdateMenus(e);

        EditorSetBusy(e, FALSE);
}


/*
 *	Check scenery callback.
 */
void EditorCheckSceneryCB(GtkWidget *widget, gpointer data)
{
	EditorDoCheck(EDITOR(data), TRUE);
}

/*
 *	Run SAR callback.
 */
void EditorRunSARCB(GtkWidget *widget, gpointer data)
{
	EditorRunSAR(EDITOR(data), TRUE);
}
