#include <glib.h>
#include "../include/fio.h"
#include "editorfio.h"
#include "config.h"


GList *EditorFIOHumanGetTypes(const gchar *path);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Loads a list of human types from the file specified by path.
 *
 *	If path is NULL then the global human.ini file will be used.
 */
GList *EditorFIOHumanGetTypes(const gchar *path)
{
	gchar *buf;
	FILE *fp;
	GList *glist = NULL;


	if(path == NULL)
	    path = SE_DEF_SAR_HUMAN_INI_FILE;

	/* Open file for reading. */
	fp = FOpen(path, "rb");
	if(fp == NULL)
	{
	    glist = g_list_append(glist, STRDUP("default"));
	    return(glist);
	}

	/* Begin reading file. */
	buf = NULL;
	while(TRUE)
	{
	    buf = FSeekNextParm(fp, buf, '#', '=');
            if(buf == NULL)
                break;

            /* Add a new preset human data entry? */
	    if(!g_strcasecmp(buf, "PresetAdd"))
            {
		gchar *s = FGetString(fp);
		glist = g_list_append(glist, s);
	    }
	}

	/* Close file. */
	FClose(fp);

	/* Ensure that we get at least one item. */
	if(glist == NULL)
            glist = g_list_append(glist, STRDUP("default"));

	return(glist);
}

