#include <stdlib.h>
#include <sys/stat.h>
#ifdef __MSW__
# include <windows.h>
#endif

#include <gtk/gtk.h>

#include "guiutils.h"
#include "cdialog.h"

#include "editor.h"
#include "editorrun.h"
#include "config.h"


gint EditorRunSAR(editor_struct *e, gboolean verbose);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Attempts to locate the SAR program and run it.
 *
 *	Returns non-zero on error.
 */
gint EditorRunSAR(editor_struct *e, gboolean verbose)
{
	gint status;
	gchar *prog, *cmd;
	GtkWidget *toplevel;
	struct stat stat_buf;


	if(e == NULL)
	    return(-1);

	toplevel = EDITOR_TOPLEVEL(e);

	/* Find SAR program. */
	prog = STRDUP(SE_DEF_SAR_PROG);
	if(stat(prog, &stat_buf))
	{
	    if(verbose)
	    {
		gchar *buf = g_strdup_printf(
"Unable to find program:\n\n    %s",
		    prog
		);
		CDialogSetTransientFor(toplevel);
		CDialogGetResponse(
		    "Run Failed",
		    buf,
		    NULL,
		    CDIALOG_ICON_ERROR,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		g_free(buf);
	    }
	    g_free(prog);
	    return(-1);
	}

	/* Format command. */
	cmd = g_strdup_printf("%s &", prog);

	/* Run command. */
	status = system(cmd);

	g_free(cmd);
	g_free(prog);

	return(status);
}
